import operator
from collections.abc import Collection
from copy import deepcopy
from itertools import product
from typing import Any, Optional


def member_of(item: Any, collection: Collection[Any]) -> bool:
    return item in collection


OPS = {
    operator.__lt__: "<",
    operator.__le__: "<=",
    operator.__eq__: "==",
    operator.__ne__: "!=",
    operator.__gt__: ">",
    operator.__ge__: ">=",
    operator.__and__: "&",
    operator.__or__: "|",
    operator.__invert__: "~",
    operator.__mod__: "%",
    # operator.__contains__: 'in',
    member_of: "in",
}


class Expression:
    """Base class for logic expressions"""

    def __eq__(self, other: "Expression"):
        if isinstance(other, Expression):
            return self._ordered_str() == other._ordered_str()
        else:
            return False

    def __and__(self, other: "Expression"):
        return And(self, other)

    def __or__(self, other: "Expression"):
        return Or(self, other)

    def __invert__(self):
        return Not(self)

    def __lt__(self, other: Any):
        return Term("<", self, other)

    def __le__(self, other: Any):
        return Term("<=", self, other)

    def __ne__(self, other: Any):
        return Term("!=", self, other)

    def __gt__(self, other: Any):
        return Term(">", self, other)

    def __ge__(self, other: Any):
        return Term(">=", self, other)

    def __mod__(self, other: Any):
        return Term("%", self, other)

    def __contains__(self, other: Any):
        return Term("contains", self, other)
        return Term("contains", self, other)

    def _ordered_str(self, **kwargs):
        return str(self)

    def __repr__(self):
        return self._ordered_str()


class Variable(Expression):
    """A variable in a logic expression"""

    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return str(self.name)

    def _ordered_str(self, **kwargs):
        return str(self.name)


class Top(Expression):
    """True expression"""

    def __str__(self):
        return "T"

    def _ordered_str(self, **kwargs):
        return str(self)


class Bottom(Expression):
    """False expression"""

    def __str__(self):
        return "F"

    def _ordered_str(self, **kwargs):
        return str(self)


class Not(Expression):
    """Negation of an expression"""

    def __init__(self, operand: Expression):
        self.operand = operand

    def __str__(self):
        return f"~{self.operand})"

    def _ordered_str(self, **kwargs):
        return str(self)


class And(Expression):
    """Conjunction of expressions"""

    def __init__(self, *operands: Expression):
        self.operands: list[Expression] = list(operands)

    def __str__(self):
        return f"({' & '.join(str(operand) for operand in self.operands)})"

    def _ordered_str(self, **kwargs):
        sorted_operands = sorted([op._ordered_str(**kwargs) for op in self.operands], key=str)
        return f"({' & '.join(sorted_operands)})"


class Or(Expression):
    def __init__(self, *operands: Expression):
        self.operands = list(operands)

    def __str__(self):
        return f"({' | '.join(str(operand) for operand in self.operands)})"

    def _ordered_str(self, pairwise=False):
        if pairwise and len(self.operands) > 2:
            return Or(Or(*self.operands[0:2]), Or(*self.operands[2:]))._ordered_str(pairwise=True)
        sorted_operands = sorted([op._ordered_str(pairwise=True) for op in self.operands], key=str)
        return f"({' | '.join(sorted_operands)})"


class Eq(Expression):
    def __init__(self, lhs: Expression, rhs: Expression):
        self.lhs = lhs
        self.rhs = rhs

    def __str__(self):
        return f"{str(self.lhs)} = {str(self.rhs)}"

    def _ordered_str(self, **kwargs):
        return str(self)


class Term(Expression):
    def __init__(self, predicate: str, *operands: Any):
        self.predicate = predicate
        self.operands = list(operands)

    def __str__(self):
        if self.predicate in OPS:
            return f"{str(self.operands[0])} {OPS[self.predicate]} {str(self.operands[1])}"
        else:
            return f"{self.predicate}({', '.join(str(operand) for operand in self.operands)})"

    def _ordered_str(self, **kwargs):
        return str(self)


class IsIn(Term):
    def __init__(self, element: Expression, collection: list[Any]):
        self.predicate = "in"
        self.operands = [element, collection]


class Value(Expression):
    # DEPRECATED
    def __init__(self, operand: Any):
        self.operand = operand

    def __str__(self):
        return str(self.operand)

    def _ordered_str(self, **kwargs):
        return str(self)


class Implies(Expression):
    def __init__(self, left: Expression, right: Expression):
        self.left = left
        self.right = right

    def __str__(self):
        return f"({self.left} => {self.right})"

    def _ordered_str(self, **kwargs):
        return f"({self.left._ordered_str(**kwargs)} => {self.right._ordered_str(**kwargs)})"


class Iff(Expression):
    def __init__(self, left: Expression, right: Expression):
        self.left = left
        self.right = right

    def __str__(self):
        return f"({self.left} => {self.right})"

    def _ordered_str(self, **kwargs):
        return f"({self.left._ordered_str(**kwargs)} iff {self.right._ordered_str(**kwargs)})"


class UniversalSet(Expression):
    def __str__(self):
        return "__U__"

    def _ordered_str(self, **kwargs):
        return str(self)


# DNF conversion functions


def eliminate_implications(expr: Expression) -> Expression:
    """
    Eliminate implications and iffs from an expression.

    - ``P => Q is equivalent to ~P | Q``
    - ``P iff Q is equivalent to (~P | Q) & (P | ~Q)``

    Also recurse through sub-expressions

    :param expr:
    :return:
    """
    if isinstance(expr, Implies):
        # P => Q is equivalent to ~P | Q
        return Or(Not(expr.left), expr.right)
    elif isinstance(expr, Iff):
        # P iff Q is equivalent to (~P | Q) & (P | ~Q)
        return And(Or(Not(expr.left), expr.right), Or(expr.left, Not(expr.right)))
    elif isinstance(expr, And) or isinstance(expr, Or):
        # Recursively eliminate implications from Ands:
        # Tr(P1 & P2 & ... & Pn) is equivalent to Tr(P1) & Tr(P2) & ... & Tr(Pn)
        expr.operands = list(map(eliminate_implications, expr.operands))
    elif isinstance(expr, Not) and isinstance(expr.operand, (And, Or, Implies, Iff)):
        # Recursively eliminate implications from Nots:
        # Tr(~(P1)) is equivalent to ~Tr(P1)
        expr.operand = eliminate_implications(expr.operand)
    return expr


def move_not_inwards(expr: Expression) -> Expression:
    """
    Move negations inwards in an expression.

    :param expr:
    :return:
    """
    if isinstance(expr, Not) and isinstance(expr.operand, Not):
        return expr.operand.operand
    elif isinstance(expr, Not) and isinstance(expr.operand, And):
        return Or(*[Not(operand) for operand in expr.operand.operands])
    elif isinstance(expr, Not) and isinstance(expr.operand, Or):
        return And(*[Not(operand) for operand in expr.operand.operands])
    elif isinstance(expr, And) or isinstance(expr, Or):
        expr.operands = list(map(move_not_inwards, expr.operands))
    elif isinstance(expr, Not) and isinstance(expr.operand, (And, Or)):
        expr.operand = move_not_inwards(expr.operand)
    return expr


def distribute_and_over_or(expr: Expression) -> Expression:
    """
    Distributes AND over OR in the given expression according to the distributive laws of Boolean Algebra.

    The function works recursively, transforming each AND operation in the expression.

    Args:
        expr (Expression): The input expression which is an instance of Expression or its subclasses.

    Returns:
        Expression: The transformed expression with AND distributed over OR.
    """
    if isinstance(expr, And):
        or_exprs = [operand for operand in expr.operands if isinstance(operand, Or)]
        other_exprs = [operand for operand in expr.operands if not isinstance(operand, Or)]
        if or_exprs:
            new_operands = []
            for combination in product(*[or_expr.operands for or_expr in or_exprs]):
                new_operands.append(And(*(other_exprs + list(combination))))
            return Or(*new_operands)
        expr.operands = list(map(distribute_and_over_or, expr.operands))
        # conj over conj
        and_exprs = [operand for operand in expr.operands if isinstance(operand, And)]
        if and_exprs:
            new_operands = []
            for and_expr in and_exprs:
                new_operands.extend(and_expr.operands)
            new_operands.extend([operand for operand in expr.operands if not isinstance(operand, And)])
            return And(*new_operands)
    elif isinstance(expr, Or):
        expr.operands = list(map(distribute_and_over_or, expr.operands))
        # disj over disj
        or_exprs = [operand for operand in expr.operands if isinstance(operand, Or)]
        if or_exprs:
            new_operands = []
            for or_expr in or_exprs:
                new_operands.extend(or_expr.operands)
            new_operands.extend([operand for operand in expr.operands if not isinstance(operand, Or)])
            return Or(*new_operands)
    elif isinstance(expr, Not):
        expr.operand = distribute_and_over_or(expr.operand)
    return expr


def is_contradiction(expr: Expression, apply_dnf=False) -> Optional[bool]:
    """
    Check if the given expression is a contradiction.

    >>> a = Variable("a")
    >>> b = Variable("b")
    >>> c = Variable("c")
    >>> d = Variable("d")
    >>> F = Bottom()
    >>> T = Top()
    >>> is_contradiction(T)
    False
    >>> is_contradiction(F)
    True
    >>> is_contradiction(~T)
    True
    >>> is_contradiction(~F)
    False
    >>> is_contradiction(~~T)
    False
    >>> is_contradiction(~~F)
    True
    >>> is_contradiction(F & F)
    True
    >>> is_contradiction(F & T)
    True
    >>> is_contradiction(F | T)
    False
    >>> is_contradiction(a & F )
    True
    >>> is_contradiction(a & ~a)
    True
    >>> assert is_contradiction(a & T) is None
    >>> assert is_contradiction(a) is None
    >>> assert is_contradiction(And(a)) is None

    :param expr:
    :param apply_dnf:
    :return:
    """
    if apply_dnf:
        expr = to_dnf(expr)
        expr = simplify(expr)
    if isinstance(expr, And):
        for i, op_i in enumerate(expr.operands):
            if isinstance(op_i, Not):
                for j, op_j in enumerate(expr.operands):
                    if op_i != op_j:
                        if op_i.operand == op_j:
                            return True
        opx = [is_contradiction(operand) for operand in expr.operands]
        if any(opx):
            return True
        elif any([x is None for x in opx]):
            return None
        else:
            return False
        # return any([is_contradiction(operand) for operand in expr.operands])
    elif isinstance(expr, Not):
        negated_expr = is_contradiction(expr.operand)
        if negated_expr is None:
            return None
        else:
            return not negated_expr
    elif isinstance(expr, Or):
        opx = [is_contradiction(operand) for operand in expr.operands]
        if all(opx):
            return True
        elif any([x is False for x in opx]):
            return False
        else:
            return None
    if expr == Bottom():
        return True
    elif expr == Top():
        return False
    elif isinstance(expr, Term) and expr.predicate == "in":
        if expr.operands[1] == UniversalSet():
            return False
        if expr.operands[1] == []:
            return True
    return None


def evals_to_true(expr: Expression, apply_dnf=False) -> Optional[bool]:
    """
    Two-valued logic

    >>> a = Variable("a")
    >>> b = Variable("b")
    >>> c = Variable("c")
    >>> d = Variable("d")
    >>> F = Bottom()
    >>> T = Top()
    >>> evals_to_true(T)
    True
    >>> evals_to_true(F)
    False
    >>> evals_to_true(~T)
    False
    >>> evals_to_true(~F)
    True
    >>> evals_to_true(a & b)
    False
    >>> evals_to_true(a | b)
    False
    >>> evals_to_true(a & a)
    False

    Note that this does not perform unification, so the following does not evaluate to True

    >>> evals_to_true(a & ~a)
    False

    :param expr:
    :param apply_dnf:
    :return:
    """
    if apply_dnf:
        expr = to_dnf(expr)
    if isinstance(expr, And):
        return all([evals_to_true(operand) for operand in expr.operands])
    elif isinstance(expr, Not):
        if is_contradiction(expr.operand) is True:
            return True
        else:
            return False
    elif isinstance(expr, Or):
        return any([evals_to_true(operand) for operand in expr.operands])
    elif expr == Top():
        return True
    elif isinstance(expr, IsIn):
        if expr.operands[1] == UniversalSet():
            return True
    else:
        return False


def eliminate_redundant(expr: Expression) -> Expression:
    if isinstance(expr, And):
        visited = set()
        new_operands = []
        for i, operand in enumerate(expr.operands):
            # if isinstance(operand, Top):
            #    continue
            operand = eliminate_redundant(operand)
            operand_str = operand._ordered_str()
            if operand_str not in visited:
                visited.add(operand_str)
                new_operands.append(operand)
        expr.operands = new_operands
    elif isinstance(expr, Or):
        visited = set()
        new_operands = []
        for i, operand in enumerate(expr.operands):
            if isinstance(operand, Bottom):
                continue
            operand = eliminate_redundant(operand)
            operand_str = operand._ordered_str()
            if operand_str not in visited:
                visited.add(operand_str)
                new_operands.append(operand)
        expr.operands = new_operands
    return expr


def simplify_full(expr: Expression) -> Expression:
    simplified = simplify(expr)
    if str(simplified) != str(expr):
        return simplify_full(simplified)
    else:
        return simplified


def simplify(expr: Expression) -> Expression:
    """
    Simplifies the given propositional logic formula.

    >>> a = Variable("a")
    >>> b = Variable("b")
    >>> c = Variable("c")
    >>> d = Variable("d")
    >>> simplify(a & b)
    (a & b)
    >>> simplify(a | b)
    (a | b)
    >>> simplify(a & a)
    a
    >>> simplify(a | a)
    a
    >>> simplify(a & ~a)
    F
    >>> simplify(a | ~a)
    T

    :param expr:
    :return:
    """
    expr = to_dnf(expr)
    if expr == ~Bottom():
        return Top()
    if expr == ~Top():
        return Bottom()
    if is_contradiction(expr, apply_dnf=False):
        return Bottom()
    if evals_to_true(expr):
        return Top()
    if isinstance(expr, (And, Or)):
        modified = True
        while modified:
            modified = False
            modifications = []
            for i, op_i in enumerate(expr.operands):
                for j, op_j in enumerate(expr.operands):
                    if i < j:
                        for composition in COMPOSITIONS[type(expr)]:
                            new_expr = composition(op_i, op_j)
                            if new_expr is not None:
                                modifications.append((i, j, new_expr))
                            else:
                                new_expr = composition(op_j, op_i)
                                if new_expr is not None:
                                    modifications.append((j, i, new_expr))
                                else:
                                    if (
                                        isinstance(op_i, Term)
                                        and isinstance(op_j, Term)
                                        and op_i.operands[0] == op_j.operands[0]
                                    ):
                                        new_tpl = compose_operators(
                                            type(expr),
                                            op_i.predicate,
                                            op_i.operands[1],
                                            op_j.predicate,
                                            op_j.operands[1],
                                        )
                                        if new_tpl is not None:
                                            new_op, new_arg = new_tpl
                                            new_expr = Term(OPS[new_op], op_i.operands[0], new_arg)
                                            modifications.append((i, j, new_expr))
            if modifications:
                modified = True
                for i, j, new_expr in modifications:
                    expr.operands[i] = new_expr
                    expr.operands[j] = None
                expr.operands = [op for op in expr.operands if op is not None]
        if len(expr.operands) == 1:
            return simplify(expr.operands[0])
        else:
            return type(expr)(*list(map(simplify, expr.operands)))
    elif isinstance(expr, Not):
        return Not(simplify(expr.operand))
    else:
        return expr


COMPOSITIONS = {
    And: [
        (lambda x, y: Bottom() if x == ~y else None),
        (lambda x, _: Bottom() if x == Bottom() else None),
        (lambda x, y: Bottom() if _unsat(x, y) else None),
        (lambda x, y: x if x == y else None),
    ],
    Or: [
        lambda x, y: Top() if x == ~y else None,
        lambda x, _: Top() if x == Top() else None,
        (lambda x, y: x if x == y else None),
    ],
}


def _unsat(x: Expression, y: Expression) -> bool:
    """
    Check if two terms are unsatisfiable when combined.

    >>> a = Variable("a")
    >>> b = Variable("b")
    >>> _unsat( a < b, a > b)
    True
    >>> _unsat( a < b, a < b)
    False
    >>> _unsat( a <= b, b >= a)
    False
    >>> _unsat( a < b, a <= b)
    False
    >>> _unsat( a < b, b <= a)
    False
    >>> _unsat( IsIn(a, [1, 2]), IsIn(a, [2, 3]))
    False
    >>> _unsat( IsIn(a, [1, 2]), IsIn(a, [3, 4]))
    True

    :param x:
    :param y:
    :return:
    """
    if isinstance(x, Term) and isinstance(y, Term):
        if x.operands[0] == y.operands[0]:
            xp = x.predicate
            yp = y.predicate
            xv = x.operands[1]
            yv = y.operands[1]
            if xp == "<" and yp == ">":
                if xv <= yv:
                    return True
            elif xp == "<" and yp == ">=":
                if xv < yv:
                    return True
            elif xp == "<=" and yp == ">":
                if xv < yv:
                    return True
            elif xp == "<=" and yp == ">=":
                if xv < yv:
                    return True
            elif xp == "in" and yp == "in":
                if xv == UniversalSet() or yv == UniversalSet():
                    return False
                if not set(xv).intersection(set(yv)):
                    return True
    return False


def compose_operators(boolean_op: type[Expression], op1: str, v1: Any, op2: str, v2: Any) -> Optional[tuple]:
    """
    Compose two expressions.

    >>> assert compose_operators(And, '<', 1, '<', 2) == (operator.__lt__, 1)
    >>> assert compose_operators(Or, '<', 1, '<', 2) == (operator.__lt__, 2)
    >>> assert compose_operators(Or, '=', 1, '=', 2) is None
    >>> assert compose_operators(And, '=', 1, '=', 2) is None
    >>> assert compose_operators(And, "in", [1, 2], "in", [2, 3]) == (member_of, [2])
    >>> assert compose_operators(Or, "in", [1, 2], "in", [2, 3]) == (member_of, [1, 2, 3])
    >>> assert compose_operators(And, "in", [1], "in", [3]) == (member_of, [])
    >>> assert compose_operators(And, "in", [1], "in", UniversalSet()) == (member_of, [1])
    >>> assert compose_operators(Or, "in", [1], "in", UniversalSet()) == (member_of, UniversalSet())

    :param boolean_op:
    :param op1:
    :param v1:
    :param op2:
    :param v2:
    :return: New expression OR None if the expressions cannot be reduced
    """
    rev_op_map = {v: k for k, v in OPS.items()}
    if not (op1 in rev_op_map and op2 in rev_op_map):
        return None
    op1 = rev_op_map[op1]
    op2 = rev_op_map[op2]
    if boolean_op == And:
        if op1 == operator.__lt__ and op2 == operator.__lt__:
            return operator.__lt__, min(v1, v2)
        elif op1 == operator.__le__ and op2 == operator.__le__:
            return operator.__le__, min(v1, v2)
        elif op1 == operator.__lt__ and op2 == operator.__le__:
            if v1 <= v2:
                return operator.__lt__, v1
            else:
                return operator.__le__, v2
        elif op1 == operator.__le__ and op2 == operator.__lt__:
            if v2 <= v1:
                return operator.__lt__, v2
            else:
                return operator.__le__, v1
        elif op1 == operator.__gt__ and op2 == operator.__gt__:
            return operator.__gt__, max(v1, v2)
        elif op1 == operator.__ge__ and op2 == operator.__ge__:
            return operator.__ge__, max(v1, v2)
        elif op1 == operator.__gt__ and op2 == operator.__ge__:
            if v1 >= v2:
                return operator.__gt__, v1
            else:
                return operator.__ge__, v2
        elif op1 == operator.__ge__ and op2 == operator.__gt__:
            if v2 >= v1:
                return operator.__gt__, v2
            else:
                return operator.__ge__, v1
        elif op1 == member_of and op2 == member_of:
            if v1 == UniversalSet():
                return member_of, v2
            elif v2 == UniversalSet():
                return member_of, v1
            else:
                return member_of, list(set(v1).intersection(v2))
    elif boolean_op == Or:
        if op1 == operator.__lt__ and op2 == operator.__lt__:
            return operator.__lt__, max(v1, v2)
        elif op1 == operator.__le__ and op2 == operator.__le__:
            return operator.__le__, max(v1, v2)
        elif op1 == operator.__lt__ and op2 == operator.__le__:
            if v1 <= v2:
                return operator.__le__, v2
            else:
                return operator.__lt__, v1
        elif op1 == operator.__le__ and op2 == operator.__lt__:
            if v2 <= v1:
                return operator.__le__, v1
            else:
                return operator.__lt__, v2
        elif op1 == operator.__gt__ and op2 == operator.__gt__:
            return operator.__gt__, min(v1, v2)
        elif op1 == operator.__ge__ and op2 == operator.__ge__:
            return operator.__ge__, min(v1, v2)
        elif op1 == operator.__gt__ and op2 == operator.__ge__:
            if v1 >= v2:
                return operator.__ge__, v2
            else:
                return operator.__gt__, v1
        elif op1 == operator.__ge__ and op2 == operator.__gt__:
            if v2 >= v1:
                return operator.__ge__, v1
            else:
                return operator.__gt__, v2
        elif op1 == member_of and op2 == member_of:
            if v1 == UniversalSet() or v2 == UniversalSet():
                return member_of, UniversalSet()
            return member_of, list(set(v1).union(v2))


def to_dnf(expr: Expression) -> Expression:
    expr = deepcopy(expr)
    expr = eliminate_implications(expr)
    expr = move_not_inwards(expr)
    expr = distribute_and_over_or(expr)
    return expr



import logging
import os
from dataclasses import dataclass
from typing import Callable

import click
from jsonasobj2 import JsonObj, as_dict
from linkml_runtime.linkml_model.meta import ClassDefinition, ElementName
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.utils.schemaview import SchemaView
from linkml_runtime.utils.yamlutils import TypedNode, extended_float, extended_int, extended_str
from rdflib import BNode, Graph, Literal, URIRef
from rdflib.collection import Collection
from rdflib.namespace import RDF, SH, XSD

from linkml._version import __version__
from linkml.generators.shacl.shacl_data_type import ShaclDataType
from linkml.generators.shacl.shacl_ifabsent_processor import ShaclIfAbsentProcessor
from linkml.utils.generator import Generator, shared_arguments

logger = logging.getLogger(__name__)


@dataclass
class ShaclGenerator(Generator):
    # ClassVars
    closed: bool = True
    """True means add 'sh:closed=true' to all shapes, except of mixin shapes and shapes, that have parents"""
    suffix: str = None
    """parameterized suffix to be appended. No suffix per default."""
    include_annotations: bool = False
    """True means include all class / slot / type annotations in generated Node or Property shapes"""
    exclude_imports: bool = False
    """If True, elements from imported ontologies won't be included in the generator's output"""
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["ttl"]
    file_extension = "shacl.ttl"
    visit_all_class_slots = False
    uses_schemaloader = True

    def __post_init__(self) -> None:
        self.schemaview = SchemaView(self.schema)
        super().__post_init__()
        self.generate_header()

    def generate_header(self) -> str:
        out = f"\n# metamodel_version: {self.schema.metamodel_version}"
        if self.schema.version:
            out += f"\n# version: {self.schema.version}"
        return out

    def serialize(self, **args) -> str:
        g = self.as_graph()
        data = g.serialize(format="turtle" if self.format in ["owl", "ttl"] else self.format)
        return data

    def as_graph(self) -> Graph:
        sv = self.schemaview
        g = Graph()
        g.bind("sh", SH)

        ifabsent_processor = ShaclIfAbsentProcessor(sv)

        for pfx in self.schema.prefixes.values():
            g.bind(str(pfx.prefix_prefix), pfx.prefix_reference)

        for c in sv.all_classes(imports=not self.exclude_imports).values():

            def shape_pv(p, v):
                if v is not None:
                    g.add((class_uri_with_suffix, p, v))

            class_uri = URIRef(sv.get_uri(c, expand=True))
            class_uri_with_suffix = class_uri
            if self.suffix is not None:
                class_uri_with_suffix += self.suffix
            shape_pv(RDF.type, SH.NodeShape)
            shape_pv(SH.targetClass, class_uri)  # TODO

            if self.closed:
                if c.mixin or c.abstract:
                    shape_pv(SH.closed, Literal(False))
                else:
                    shape_pv(SH.closed, Literal(True))
            else:
                shape_pv(SH.closed, Literal(False))
            if c.title is not None:
                shape_pv(SH.name, Literal(c.title))
            if c.description is not None:
                shape_pv(SH.description, Literal(c.description))

            shape_pv(SH.ignoredProperties, self._build_ignored_properties(g, c))

            if c.annotations and self.include_annotations:
                self._add_annotations(shape_pv, c)
            order = 0
            for s in sv.class_induced_slots(c.name):
                # fixed in linkml-runtime 1.1.3
                if s.name in sv.element_by_schema_map():
                    slot_uri = URIRef(sv.get_uri(s, expand=True))
                else:
                    pfx = sv.schema.default_prefix
                    slot_uri = URIRef(sv.expand_curie(f"{pfx}:{underscore(s.name)}"))
                pnode = BNode()
                shape_pv(SH.property, pnode)

                def prop_pv(p, v):
                    if v is not None:
                        g.add((pnode, p, v))

                def prop_pv_literal(p, v):
                    if v is not None:
                        g.add((pnode, p, Literal(v)))

                prop_pv(SH.path, slot_uri)
                prop_pv_literal(SH.order, order)
                order += 1
                prop_pv_literal(SH.name, s.title)
                prop_pv_literal(SH.description, s.description)
                # minCount
                if s.minimum_cardinality:
                    prop_pv_literal(SH.minCount, s.minimum_cardinality)
                elif s.exact_cardinality:
                    prop_pv_literal(SH.minCount, s.exact_cardinality)
                elif s.required:
                    prop_pv_literal(SH.minCount, 1)
                # maxCount
                if s.maximum_cardinality:
                    prop_pv_literal(SH.maxCount, s.maximum_cardinality)
                elif s.exact_cardinality:
                    prop_pv_literal(SH.maxCount, s.exact_cardinality)
                elif not s.multivalued:
                    prop_pv_literal(SH.maxCount, 1)
                prop_pv_literal(SH.minInclusive, s.minimum_value)
                prop_pv_literal(SH.maxInclusive, s.maximum_value)

                all_classes = sv.all_classes()
                if s.any_of:
                    # It is not allowed to use any of and equals_string or equals_string_in in one
                    # slot definition, as both are mapped to sh:in in SHACL
                    if s.equals_string or s.equals_string_in:
                        error = "'equals_string'/'equals_string_in' and 'any_of' are mutually exclusive"
                        raise ValueError(f"{TypedNode.yaml_loc(str(s), suffix='')} {error}")

                    or_node = BNode()
                    prop_pv(SH["or"], or_node)
                    range_list = []
                    for any in s.any_of:
                        r = any.range
                        if r in all_classes:
                            class_node = BNode()

                            def cl_node_pv(p, v):
                                if v is not None:
                                    g.add((class_node, p, v))

                            self._add_class(cl_node_pv, r)
                            range_list.append(class_node)
                        elif r in sv.all_types():
                            t_node = BNode()

                            def t_node_pv(p, v):
                                if v is not None:
                                    g.add((t_node, p, v))

                            self._add_type(t_node_pv, r)
                            range_list.append(t_node)
                        elif r in sv.all_enums():
                            en_node = BNode()

                            def en_node_pv(p, v):
                                if v is not None:
                                    g.add((en_node, p, v))

                            self._add_enum(g, en_node_pv, r)
                            range_list.append(en_node)
                        else:
                            st_node = BNode()

                            def st_node_pv(p, v):
                                if v is not None:
                                    g.add((st_node, p, v))

                            add_simple_data_type(st_node_pv, r)
                            range_list.append(st_node)
                    Collection(g, or_node, range_list)
                else:
                    prop_pv_literal(SH.hasValue, s.equals_number)
                    r = s.range
                    if s.equals_string or s.equals_string_in:
                        # Check if range is "string" as this is mandatory for "equals_string" and "equals_string_in"
                        if r != "string":
                            raise ValueError(
                                f"slot: \"{slot_uri}\" - 'equals_string' and 'equals_string_in'"
                                f" require range 'string' and not '{r}'"
                            )

                    if r in all_classes:
                        self._add_class(prop_pv, r)
                        if sv.get_identifier_slot(r) is not None:
                            prop_pv(SH.nodeKind, SH.IRI)
                        else:
                            prop_pv(SH.nodeKind, SH.BlankNodeOrIRI)
                    elif r in sv.all_types():
                        self._add_type(prop_pv, r)
                    elif r in sv.all_enums():
                        self._add_enum(g, prop_pv, r)
                    else:
                        add_simple_data_type(prop_pv, r)
                    if s.pattern:
                        prop_pv(SH.pattern, Literal(s.pattern))
                    if s.equals_string:
                        # Map equal_string and equal_string_in to sh:in
                        self._and_equals_string(g, prop_pv, [s.equals_string])
                    if s.equals_string_in:
                        # Map equal_string and equal_string_in to sh:in
                        self._and_equals_string(g, prop_pv, s.equals_string_in)

                if s.annotations and self.include_annotations:
                    self._add_annotations(prop_pv, s)

                default_value = ifabsent_processor.process_slot(s, c)
                if default_value:
                    prop_pv(SH.defaultValue, default_value)

        return g

    def _add_class(self, func: Callable, r: ElementName) -> None:
        sv = self.schemaview
        range_ref = sv.get_uri(r, expand=True)
        func(SH["class"], URIRef(range_ref))

    def _add_enum(self, g: Graph, func: Callable, r: ElementName) -> None:
        sv = self.schemaview
        enum = sv.get_enum(r)
        pv_node = BNode()
        Collection(
            g,
            pv_node,
            [
                URIRef(sv.expand_curie(pv.meaning)) if pv.meaning else Literal(pv_name)
                for pv_name, pv in enum.permissible_values.items()
            ],
        )
        func(SH["in"], pv_node)

    def _add_type(self, func: Callable, r: ElementName) -> None:
        func(SH.nodeKind, SH.Literal)
        sv = self.schemaview
        rt = sv.get_type(r)
        if rt.uri:
            func(SH.datatype, URIRef(sv.get_uri(rt, expand=True)))
            if rt.pattern:
                func(SH.pattern, Literal(rt.pattern))
            if rt.annotations and self.include_annotations:
                self._add_annotations(func, rt)
        else:
            logger.error(f"No URI for type {rt.name}")

    def _and_equals_string(self, g: Graph, func: Callable, values: list) -> None:
        pv_node = BNode()
        Collection(
            g,
            pv_node,
            [Literal(v) for v in values],
        )
        func(SH["in"], pv_node)

    def _add_annotations(self, func: Callable, item) -> None:
        # TODO: migrate some of this logic to SchemaView
        sv = self.schemaview
        annotations = item.annotations
        # item could be a class, slot or type
        # annotation type could be dict (on types) or JsonObj (on slots)
        if type(annotations) is JsonObj:
            annotations = as_dict(annotations)
        for a in annotations.values():
            # If ':' is in the tag, treat it as a CURIE, otherwise string Literal
            if ":" in a["tag"]:
                N_predicate = URIRef(sv.expand_curie(a["tag"]))
            else:
                N_predicate = Literal(a["tag"], datatype=XSD.string)
            # If the value is a string and ':' is in the value, treat it as a CURIE,
            # otherwise treat as Literal with derived XSD datatype
            if type(a["value"]) is extended_str and ":" in a["value"]:
                N_object = URIRef(sv.expand_curie(a["value"]))
            else:
                N_object = Literal(a["value"], datatype=self._getXSDtype(a["value"]))

            func(N_predicate, N_object)

    def _getXSDtype(self, value):
        value_type = type(value)
        if value_type is bool:
            return XSD.boolean
        elif value_type is extended_str:
            return XSD.string
        elif value_type is extended_int:
            return XSD.integer
        elif value_type is extended_float:
            # TODO: distinguish between xsd:decimal and xsd:double?
            return XSD.decimal
        else:
            return None

    def _and_equals_string(self, g: Graph, func: Callable, values: list) -> None:
        pv_node = BNode()
        Collection(
            g,
            pv_node,
            [Literal(v) for v in values],
        )
        func(SH["in"], pv_node)

    def _build_ignored_properties(self, g: Graph, c: ClassDefinition) -> BNode:
        def collect_child_properties(class_name: str, output: set) -> None:
            for childName in self.schemaview.class_children(class_name, imports=True, mixins=False, is_a=True):
                output.update(
                    {
                        URIRef(self.schemaview.get_uri(prop, expand=True))
                        for prop in self.schemaview.class_slots(childName)
                    }
                )
                collect_child_properties(childName, output)

        child_properties = set()
        collect_child_properties(c.name, child_properties)

        class_slot_uris = {
            URIRef(self.schemaview.get_uri(prop, expand=True)) for prop in self.schemaview.class_slots(c.name)
        }
        ignored_properties = child_properties.difference(class_slot_uris)

        list_node = BNode()
        ignored_properties.add(RDF.type)
        Collection(g, list_node, list(ignored_properties))

        return list_node


def add_simple_data_type(func: Callable, r: ElementName) -> None:
    for datatype in list(ShaclDataType):
        if datatype.linkml_type == r:
            func(SH.datatype, datatype.uri_ref)


@shared_arguments(ShaclGenerator)
@click.command(name="shacl")
@click.option(
    "--closed/--non-closed",
    default=True,
    show_default=True,
    help="Use '--closed' to generate closed SHACL shapes. Use '--non-closed' to generate open SHACL shapes.",
)
@click.option(
    "-s",
    "--suffix",
    default=None,
    show_default=True,
    help="Use --suffix to append given string to SHACL class name (e. g. --suffix Shape: Person becomes PersonShape).",
)
@click.option(
    "--include-annotations/--exclude-annotations",
    default=False,
    show_default=True,
    help="Use --include-annotations to include annotations of slots, types, and classes in the generated SHACL shapes.",
)
@click.option(
    "--exclude-imports/--include-imports",
    default=False,
    show_default=True,
    help="Use --exclude-imports to exclude imported elements from the generated SHACL shapes. This is useful when "
    "extending a substantial ontology to avoid large output files.",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **args):
    """Generate SHACL turtle from a LinkML model"""
    gen = ShaclGenerator(yamlfile, **args)
    print(gen.serialize())


if __name__ == "__main__":
    cli()



import re
from functools import lru_cache
from typing import Union

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model.meta import (
    ClassDefinition,
    ElementName,
    SlotDefinition,
)


def remove_duplicates(lst):
    """Remove duplicate tuples from a list of tuples."""
    return [t for t in (set(tuple(i) for i in lst))]


def write_to_file(file_path, data, mode="w", encoding="utf-8"):
    with open(file_path, mode, encoding=encoding) as f:
        f.write(data)


def convert_to_snake_case(str):
    str = re.sub(r"(?<=[a-z])(?=[A-Z])|[^a-zA-Z]", " ", str).strip().replace(" ", "_")
    return "".join(str.lower())


@lru_cache(None)
def get_range_associated_slots(
    schemaview: SchemaView, range_class: ClassDefinition
) -> tuple[Union[SlotDefinition, None], Union[SlotDefinition, None], Union[list[SlotDefinition], None]]:
    if isinstance(range_class, ElementName):
        range_class = schemaview.get_class(range_class)
    if range_class is None:
        return None, None, None

    range_class_id_slot = schemaview.get_identifier_slot(range_class.name, use_key=True)
    if range_class_id_slot is None:
        return None, None, None

    non_id_slots = [s for s in schemaview.class_induced_slots(range_class.name) if s.name != range_class_id_slot.name]
    non_id_required_slots = [s for s in non_id_slots if s.required]

    # Some lists of objects can be serialized as SimpleDicts.
    # A SimpleDict is serialized as simple key-value pairs where the value is atomic.
    # The key must be declared as a key, and the value must satisfy one of the following conditions:
    # 1. The value slot is the only other slot in the object other than the key
    # 2. The value slot is explicitly annotated as a simple_dict_value
    # 3. The value slot is the only non-key that is required
    # See also: https://github.com/linkml/linkml/issues/1250
    range_simple_dict_value_slot = None
    if len(non_id_slots) == 1:
        range_simple_dict_value_slot = non_id_slots[0]
    elif len(non_id_slots) > 1:
        candidate_non_id_slots = []
        for non_id_slot in non_id_slots:
            if isinstance(non_id_slot.annotations, dict):
                is_simple_dict_value = non_id_slot.annotations.get("simple_dict_value", False)
            else:
                is_simple_dict_value = getattr(non_id_slot.annotations, "simple_dict_value", False)
            if is_simple_dict_value:
                candidate_non_id_slots.append(non_id_slot)
        if len(candidate_non_id_slots) == 1:
            range_simple_dict_value_slot = candidate_non_id_slots[0]
        else:
            candidate_non_id_slots = []
            for non_id_slot in non_id_slots:
                if non_id_slot.required:
                    candidate_non_id_slots.append(non_id_slot)
            if len(candidate_non_id_slots) == 1:
                range_simple_dict_value_slot = candidate_non_id_slots[0]

    return range_class_id_slot, range_simple_dict_value_slot, non_id_required_slots


def is_simple_dict(schemaview: SchemaView, slot: SlotDefinition) -> bool:
    if not slot.multivalued or not slot.inlined or slot.inlined_as_list:
        return False
    else:
        _, range_simple_dict_value_slot, _ = get_range_associated_slots(schemaview, slot.range)
        return range_simple_dict_value_slot is not None



"""
YAML Schema to RDF Generator

Generate a JSON LD representation of the model

"""

import os
import urllib.parse as urlparse
from copy import deepcopy
from dataclasses import dataclass
from typing import Optional

import click
from linkml_runtime.linkml_model import SchemaDefinition
from rdflib import Graph
from rdflib.plugin import Parser as rdflib_Parser
from rdflib.plugin import plugins as rdflib_plugins

from linkml import METAMODEL_CONTEXT_URI
from linkml._version import __version__
from linkml.generators.jsonldgen import JSONLDGenerator
from linkml.utils.generator import Generator, shared_arguments


@dataclass
class RDFGenerator(Generator):
    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    # TODO: we leave ttl as default for backwards compatibility but nt is
    # recommended, see https://github.com/linkml/linkml/issues/163
    valid_formats = ["ttl"] + sorted([x.name for x in rdflib_plugins(None, rdflib_Parser) if "/" not in str(x.name)])
    visit_all_class_slots = False
    uses_schemaloader = True

    # ObjectVars
    emit_metadata: bool = False
    context: list[str] = None
    original_schema: SchemaDefinition = None
    """See https://github.com/linkml/linkml/issues/871"""

    def __post_init__(self):
        self.original_schema = deepcopy(self.schema)
        super().__post_init__()

    def _data(self, g: Graph) -> str:
        return g.serialize(format="turtle" if self.format == "ttl" else self.format)

    def end_schema(self, output: Optional[str] = None, context: str = None, **_) -> str:
        gen = JSONLDGenerator(
            self.original_schema,
            format=JSONLDGenerator.valid_formats[0],
            metadata=self.emit_metadata,
            importmap=self.importmap,
        )
        # Iterate over permissible text strings making them URI compatible
        for e in gen.schema.enums.values():
            for pv in e.permissible_values.values():
                pv.text = urlparse.quote(pv.text)
        jsonld_str = gen.serialize(context=context)

        graph = Graph()
        graph.parse(
            data=jsonld_str,
            format="json-ld",
            base=str(self.namespaces._base),
            prefix=True,
        )
        if output:
            # Binary-safe when -o/--output is used:
            # delegate to RDFLib (Graph.serialize(destination=..., format=...)).
            # Serializers that produce bytes write directly to the file; stdout stays empty.
            fmt = "turtle" if self.format == "ttl" else self.format
            try:
                out = graph.serialize(format=fmt)
            except UnicodeDecodeError:
                graph.serialize(destination=output, format=fmt)
                return ""
            with open(output, "w", encoding="UTF-8") as outf:
                outf.write(out)
            return out

        return self._data(graph)


@shared_arguments(RDFGenerator)
@click.command(name="rdf")
@click.option("-o", "--output", help="Output file name")
@click.option(
    "--context",
    default=[METAMODEL_CONTEXT_URI],
    show_default=True,
    multiple=True,
    help=f"JSONLD context file (default: {METAMODEL_CONTEXT_URI})",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, **kwargs):
    """Generate an RDF representation of a LinkML model"""
    print(RDFGenerator(yamlfile, **kwargs).serialize(**kwargs))


if __name__ == "__main__":
    cli()



from enum import Enum

from rdflib import URIRef


class DataType:
    linkml_type: str
    uri_ref: URIRef


class ShaclDataType(DataType, Enum):
    STRING = ("string", URIRef("http://www.w3.org/2001/XMLSchema#string"))
    BOOLEAN = ("boolean", URIRef("http://www.w3.org/2001/XMLSchema#boolean"))
    DURATION = ("duration", URIRef("http://www.w3.org/2001/XMLSchema#duration"))
    DATETIME = ("datetime", URIRef("http://www.w3.org/2001/XMLSchema#dateTime"))
    DATE = ("date", URIRef("http://www.w3.org/2001/XMLSchema#date"))
    TIME = ("time", URIRef("http://www.w3.org/2001/XMLSchema#time"))
    DECIMAL = ("decimal", URIRef("http://www.w3.org/2001/XMLSchema#decimal"))
    INTEGER = ("integer", URIRef("http://www.w3.org/2001/XMLSchema#integer"))
    FLOAT = ("float", URIRef("http://www.w3.org/2001/XMLSchema#float"))
    DOUBLE = ("double", URIRef("http://www.w3.org/2001/XMLSchema#double"))
    URI = ("uri", URIRef("http://www.w3.org/2001/XMLSchema#anyURI"))
    CURIE = ("curi", URIRef("http://www.w3.org/2001/XMLSchema#string"))
    NCNAME = ("ncname", URIRef("http://www.w3.org/2001/XMLSchema#string"))
    OBJECT_IDENTIFIER = ("objectidentifier", URIRef("http://www.w3.org/ns/shex#iri"))
    NODE_IDENTIFIER = ("nodeidentifier", URIRef("http://www.w3.org/ns/shex#nonLiteral"))
    JSON_POINTER = ("jsonpointer", URIRef("http://www.w3.org/2001/XMLSchema#string"))
    JSON_PATH = ("jsonpath", URIRef("http://www.w3.org/2001/XMLSchema#string"))
    SPARQL_PATH = ("sparqlpath", URIRef("http://www.w3.org/2001/XMLSchema#string"))

    def __new__(cls, linkml_type, uri_ref):
        obj = object.__new__(cls)
        obj.linkml_type = linkml_type
        obj.uri_ref = uri_ref

        return obj

    def __init__(self, linkml_type, uri_ref):
        self.linkml_type = linkml_type
        self.uri_ref = uri_ref



import json
from typing import TextIO, Union

import jsonschema
from linkml_runtime.dumpers import json_dumper
from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.utils.yamlutils import YAMLRoot

from linkml.generators.jsonschemagen import JsonSchemaGenerator
from linkml.utils.deprecation import deprecation_warning


def _as_dict(inst):
    # TODO: replace this with linkml_runtime.dictutils when 1.0.14 is released
    inst_dict = json.loads(json_dumper.dumps(element=inst))
    del inst_dict["@type"]
    return inst_dict


def validate_object(
    data: YAMLRoot,
    schema: Union[str, TextIO, SchemaDefinition],
    target_class: type[YAMLRoot] = None,
    closed: bool = True,
):
    """
    validates instance data against a schema

    :param data: LinkML instance to be validates
    :param schema: LinkML schema
    :param target_class: class in schema to validate against
    :param closed:
    :return:
    """
    deprecation_warning("validators")
    if target_class is None:
        target_class = type(data)
    inst_dict = _as_dict(data)
    not_closed = not closed
    jsonschemastr = JsonSchemaGenerator(
        schema,
        mergeimports=True,
        top_class=target_class.class_name,
        not_closed=not_closed,
    ).serialize(not_closed=not_closed)
    jsonschema_obj = json.loads(jsonschemastr)
    return jsonschema.validate(
        inst_dict, schema=jsonschema_obj, format_checker=jsonschema.Draft7Validator.FORMAT_CHECKER
    )



from linkml.generators.oocodegen import OODocument


class DataframeDocument(OODocument):
    """Serves as an adapter between the template that renders the form of the
    dataframe schema and the LinkML model and schema view.

    Currently a thin wrapper around OODocument
    until the dataframe requirements are fully understood.
    """



import logging
import os
from dataclasses import dataclass
from typing import Optional

import click
from jinja2 import Template
from linkml_runtime.linkml_model.meta import ClassDefinition, ClassDefinitionName, Element, SlotDefinition
from linkml_runtime.utils.formatutils import camelcase, underscore

from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

logger = logging.getLogger(__name__)

type_map = {
    "str": "string",
    "int": "int",
    "Bool": "bool",
    "float": "float64",
    "XSDDate": "time.Date",
}

default_template = """
{%- if '_' in view.schema.name -%}
    {%- set package_name = view.schema.name[:view.schema.name.find('_')] -%}
{%- else -%}
    {%- set package_name = view.schema.name -%}
{%- endif -%}
package {{package_name}}

{% for c in view.all_classes().values() -%}
    {%- for sn in view.class_slots(c.name, direct=False) %}
        {%- set s = view.induced_slot(sn, c.name) -%}
        {%- if "time." in gen.range(s) -%}
            {%- set usesTime = True %}
        {%- else -%}
            {%- set usesTime = False %}
        {%- endif -%}
    {%- endfor -%}
{%- endfor -%}
{%- if usesTime -%}
import (
    "time" // for time.Date
)
{%- endif -%}

{% for c in view.all_classes().values() -%}
{%- if c.description -%}
/*
 * {{c.description}}
 */
{%- endif -%}
{% set parents = gen.parents(c) %}
type {{gen.name(c)}} struct {
    {%- if parents %}
	/*
	 * parent types
	 */
    {%- for p in parents %}
	{{p}}
    {%- endfor %}
    {%- endif -%}
    {%- for sn in view.class_slots(c.name, direct=False) %}
    {%- set s = view.induced_slot(sn, c.name) -%}
    {%- if s.description %}
	/*
	 * {{s.description}}
	 */
    {%- endif %}
	{{gen.name(s)}} {{gen.range(s)}} `json:"{{gen.json_name(s)}}"`
    {%- endfor %}
}

{% endfor %}
"""  # noqa: E101, W191


@dataclass
class GolangGenerator(Generator):
    """
    Generates Golang code from a schema
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.0"
    valid_formats = ["text"]
    uses_schemaloader = False

    def serialize(self) -> str:
        """
        Serialize a schema to Golang string
        :return:
        """
        template_obj = Template(default_template)
        out_str = template_obj.render(gen=self, schema=self.schemaview.schema, view=self.schemaview)
        return out_str

    @staticmethod
    def name(element: Element) -> str:
        """
        Returns the name of the element in its canonical form

        :param element:
        :return:
        """
        alias = element.name
        if isinstance(element, SlotDefinition) and element.alias:
            alias = element.alias
        return camelcase(alias)

    @staticmethod
    def json_name(element: Element) -> str:
        """
        Returns the name of the element in its JSON (snake-case) form

        :param element:
        :return:
        """
        alias = element.name
        if isinstance(element, SlotDefinition) and element.alias:
            alias = element.alias
        return underscore(alias)

    def classref(self, cls: ClassDefinition) -> Optional[str]:
        """
        Returns the class name for the class that holds a reference (foreign key) to members of this class

        E.g. if a class Person has an identifier field called unique_id, then this will
        return PersonUniqueId

        :param cls:
        :return: ref name, None if no identifier
        """
        id_slot = self.get_identifier_or_key_slot(cls.name)
        if id_slot:
            return f"{self.name(cls)}{camelcase(id_slot.name)}"
        else:
            return None

    def get_identifier_or_key_slot(self, cn: ClassDefinitionName) -> Optional[SlotDefinition]:
        sv = self.schemaview
        id_slot = sv.get_identifier_slot(cn)
        if id_slot:
            return id_slot
        else:
            for s in sv.class_induced_slots(cn):
                if s.key:
                    return s
            return None

    def range(self, slot: SlotDefinition) -> str:
        sv = self.schemaview
        r = slot.range
        if r in sv.all_classes():
            rc = sv.get_class(r)
            rc_ref = self.classref(rc)
            rc_name = self.name(rc)
            id_slot = self.get_identifier_or_key_slot(r)
            if slot.multivalued:
                if not id_slot or slot.inlined or slot.inlined_as_list:
                    if slot.inlined_as_list or not id_slot:
                        return f"[]{rc_name}"
                    else:
                        return f"[]{rc_name}"
                else:
                    return f"{rc_ref}[]"
            else:
                if not id_slot or slot.inlined:
                    return rc_name
                else:
                    return f"{rc_ref}"
        else:
            if r in sv.all_types():
                t = sv.get_type(r)
                if t.base and t.base in type_map:
                    return type_map[t.base]
                else:
                    logger.warning(f"Unknown type.base: {t.name}")
            return "string"

    @staticmethod
    def parents(cls: ClassDefinition) -> list[ClassDefinitionName]:
        if cls.is_a:
            parents = [cls.is_a]
        else:
            parents = []
        return [ClassDefinitionName(camelcase(p)) for p in parents + cls.mixins]


@shared_arguments(GolangGenerator)
@click.version_option(__version__, "-V", "--version")
@click.command(name="golang")
def cli(yamlfile, **args):
    """Generate Golang types

    This very simple generator produces a Golang package named after the given
    schema with structs that implement the classes in that schema.
    """
    gen = GolangGenerator(yamlfile, **args)
    print(gen.serialize())


if __name__ == "__main__":
    cli()



"""Generate OWL ontology representation of a LinkML schema."""

import logging
import os
from collections import defaultdict
from collections.abc import Mapping
from copy import copy
from dataclasses import dataclass, field
from enum import Enum, unique
from typing import Any, Optional, Union

import click
import rdflib
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model.meta import (
    AnonymousClassExpression,
    AnonymousSlotExpression,
    AnonymousTypeExpression,
    ClassDefinition,
    ClassDefinitionName,
    ClassRule,
    Definition,
    EnumDefinition,
    EnumDefinitionName,
    PermissibleValue,
    SchemaDefinitionName,
    SlotDefinition,
    SlotDefinitionName,
    TypeDefinition,
    TypeDefinitionName,
)
from linkml_runtime.utils.formatutils import camelcase, underscore
from linkml_runtime.utils.introspection import package_schemaview
from rdflib import DCTERMS, OWL, RDF, XSD, BNode, Graph, Literal, URIRef
from rdflib.collection import Collection
from rdflib.namespace import RDFS, SKOS
from rdflib.plugin import Parser as rdflib_Parser
from rdflib.plugin import plugins as rdflib_plugins

from linkml import METAMODEL_NAMESPACE_NAME
from linkml._version import __version__
from linkml.utils.generator import Generator, shared_arguments

logger = logging.getLogger(__name__)

OWL_TYPE = URIRef  ## RDFS.Literal or OWL.Thing

SWRL = rdflib.Namespace("http://www.w3.org/2003/11/swrl#")
SWRLB = rdflib.Namespace("http://www.w3.org/2003/11/swrlb#")


@unique
class MetadataProfile(Enum):
    """
    An enumeration of the different kinds of profiles used for
    metadata of generated OWL elements
    """

    linkml = "linkml"
    """Default, uses the slot URIs from the LinkML metamodel"""

    rdfs = "rdfs"
    """RDFS conventions, using core RDFS properties preferentially"""

    ols = "ols"

    @staticmethod
    def list():
        return list(map(lambda c: c.value, MetadataProfile))


@unique
class OWLProfile(Enum):
    """
    An enumeration of OWL Profiles.
    """

    dl = "dl"
    """Here this means strict OWL DL, with no punning."""

    full = "full"
    """May include punning (metaclasses)."""

    @staticmethod
    def list():
        return list(map(lambda c: c.value, MetadataProfile))


@dataclass
class OwlSchemaGenerator(Generator):
    """
    Generates a schema-oriented OWL representation of a LinkML model

    `OWL Generator Docs <https://linkml.io/linkml/generators/owl>`_

    .

    - LinkML ClassDefinitions are translated to OWL Classes
    - LinkML SlotDefinitions are translated to OWL Properties
    - LinkML Enumerations are translated to OWL Classes
    - LinkML TypeDefinitions are translated to OWL Datatypes

    The translation aims to be as faithful as possible. But note that OWL is open-world,
    whereas LinkML is closed-world
    """

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorversion = "0.1.1"
    valid_formats = ["owl", "ttl"] + [x.name for x in rdflib_plugins(None, rdflib_Parser) if "/" not in str(x.name)]
    file_extension = "owl"
    uses_schemaloader = False

    ontology_uri_suffix: str = None
    """Suffix to add to the schema name to create the ontology URI, e.g. .owl.ttl"""

    # ObjectVars
    metadata_profile: MetadataProfile = None
    """Deprecated - use metadata_profiles."""

    metadata_profiles: list[MetadataProfile] = field(default_factory=lambda: [])
    """By default, use the linkml metadata profile,
    this allows for overrides."""

    metaclasses: bool = True
    """if True, include OWL representations of ClassDefinition, SlotDefinition, etc. Introduces punning"""

    add_root_classes: bool = False

    add_ols_annotations: bool = True
    graph: Optional[Graph] = None
    """Mutable graph that is being built up during OWL generation."""

    top_value_uri: Optional[URIRef] = None
    """If metaclasses=True, then this property is used to connect object shadows to literals"""

    type_objects: bool = True
    """if True, represents types as classes (and thus all slots are object properties);
    typed object classes effectively shadow the main xsd literal types.
    The purpose of this is to allow a uniform ObjectProperty representation for all slots,
    without having to commit to being either Data or Object property (OWL-DL does not
    allow a property to be both."""

    assert_equivalent_classes: bool = False
    """If True, assert equivalence between definition_uris and class_uris"""

    use_native_uris: bool = True
    """If True, use the definition_uris, otherwise use class_uris."""

    mixins_as_expressions: bool = None
    """EXPERIMENTAL: If True, use OWL existential restrictions to represent mixins"""

    default_permissible_value_type: Union[str, URIRef] = field(default_factory=lambda: OWL.Class)

    slot_is_literal_map: Mapping[str, set[bool]] = field(default_factory=lambda: defaultdict(set))
    """DEPRECATED: use node_owltypes"""

    node_owltypes: Mapping[Union[BNode, URIRef], set[OWL_TYPE]] = field(default_factory=lambda: defaultdict(set))
    """rdfs:Datatype, owl:Thing"""

    simplify: bool = True
    """Reduce complex expressions to simpler forms"""

    use_swrl: bool = False
    """Use of SWRL is experimental"""

    target_profile: OWLProfile = field(default_factory=lambda: OWLProfile.dl)
    """Target OWL profile. Currently the only distinction is between DL and Full"""

    metamodel_schemaview: SchemaView = field(
        default_factory=lambda: package_schemaview("linkml_runtime.linkml_model.meta")
    )

    enum_iri_separator: str = "#"
    """Separator for enum IRI. Can be overridden for example if your namespace IRI already contains a #"""

    def as_graph(self) -> Graph:
        """
        Generate an rdflib Graph from the LinkML schema.

        :return:
        """
        sv = self.schemaview
        schema = sv.schema
        owl_id = schema.id
        if self.ontology_uri_suffix:
            owl_id = f"{owl_id}{self.ontology_uri_suffix}"
        mergeimports = self.mergeimports
        base = URIRef(owl_id)
        # initialize the rdflib Graph where all axiom triples will be added
        graph = Graph(identifier=base)
        self.graph = graph
        for prefix in self.metamodel.schema.emit_prefixes:
            self.graph.bind(prefix, self.metamodel.namespaces[prefix])
        for pfx in schema.prefixes.values():
            self.graph.namespace_manager.bind(pfx.prefix_prefix, URIRef(pfx.prefix_reference))
        graph.add((base, RDF.type, OWL.Ontology))

        # Add main schema elements
        for cls in sv.all_classes(imports=mergeimports).values():
            self.add_class(cls)
            for a in cls.attributes.values():
                self.add_slot(a, attribute=True)
        for slot in sv.all_slots(imports=mergeimports, attributes=False).values():
            self.add_slot(slot, attribute=False)
        for typ in sv.all_types(imports=mergeimports).values():
            self.add_type(typ)
        for enm in sv.all_enums(imports=mergeimports).values():
            self.add_enum(enm)

        if not mergeimports:
            for imp in schema.imports:
                if imp == "linkml:types":
                    continue
                graph.add((base, OWL.imports, self._schema_uri(imp)))

        # Add metadata as annotation properties
        self.add_metadata(schema, base)
        return graph

    def serialize(self, **kwargs) -> str:
        """
        Serialize the OWL triple graph to a standard RDF serialization format.

        :param kwargs:
        :return:
        """
        self.as_graph()
        data = self.graph.serialize(format="turtle" if self.format in ["owl", "ttl"] else self.format)
        return data

    def add_metadata(self, e: Union[Definition, PermissibleValue], uri: URIRef) -> None:
        """
        Add annotation properties.

        Set the profile attribute to the appropriate OWL profile.

        :param e: schema element
        :param uri: URI representation of schema element
        :return:
        """

        msv = self.metamodel_schemaview
        this_sv = self.schemaview
        sn_mappings = msv.slot_name_mappings()

        # iterate through all the assigned metamodel slots
        for metaslot_name, metaslot_value in vars(e).items():
            if not metaslot_value:
                # ignore if unset or an empty list
                continue
            metaslot_name = sn_mappings.get(metaslot_name).name
            metaslot = msv.induced_slot(metaslot_name, e.class_name)
            metaslot_curie = msv.get_uri(metaslot, native=False, expand=False)
            if metaslot_curie.startswith("linkml:"):
                # only mapped properties
                continue
            metaslot_uri = URIRef(msv.get_uri(metaslot, native=False, expand=True))
            if metaslot_name == "description" and self.has_profile(MetadataProfile.rdfs):
                metaslot_uri = RDFS.comment
            metaslot_range = metaslot.range
            if not isinstance(metaslot_value, list):
                metaslot_value = [metaslot_value]
            for v in metaslot_value:
                if metaslot_range in msv.all_types():
                    if metaslot_range == "uri":
                        obj = URIRef(v)
                    elif metaslot_range == "uriorcurie":
                        obj = URIRef(this_sv.expand_curie(v))
                    else:
                        obj = Literal(v)
                elif metaslot_range in msv.all_subsets():
                    obj = Literal(v)  # TODO
                elif metaslot_range in msv.all_classes():
                    continue
                    # if isinstance(v, str):
                    #    obj = URIRef(msv.expand_curie(v))
                    # else:
                    #    logger.debug(f"Skipping {uri} {metaslot_uri} => {v}")
                else:
                    obj = Literal(v)
                self.graph.add((uri, metaslot_uri, obj))

        for k, v in e.annotations.items():
            if isinstance(v, dict) or isinstance(v, list):
                continue
            if ":" not in k:
                default_prefix = this_sv.schema.default_prefix
                if default_prefix in this_sv.schema.prefixes:
                    default_prefix = this_sv.schema.prefixes[default_prefix].prefix_reference
                k = default_prefix + k
                k_uri = this_sv.expand_curie(k)
            else:
                k_uri = this_sv.expand_curie(k)
                if k_uri == k:
                    k_uri = None
            if k_uri:
                self.graph.add((uri, URIRef(k_uri), Literal(v.value)))

    def add_class(self, cls: ClassDefinition) -> None:
        """
        Each ClassDefinition is represented as an OWL class.

        * the OWL Class will instantiate ClassDefinition, if schema.metaclasses is true
        * the OWL Class will be annotated using the same properties as the source ClassDefinition
        * induced slots and their ranges added as OWL restrictions; note this will be under the Open World Assumption

        This method works by generating an OWL ontology via populating triples in a graph.
        To understand how the RDF-level operations here related to the OWL
        representation, consult https://www.w3.org/TR/owl2-mapping-to-rdf/

        :param cls:
        :return:
        """
        sv = self.schemaview
        cls_uri = self._class_uri(cls.name)
        self.add_metadata(cls, cls_uri)
        # add declaration
        self.graph.add((cls_uri, RDF.type, OWL.Class))
        if self.metaclasses:
            # instantiate metaclasses -- introduces punning
            self.graph.add(
                (
                    cls_uri,
                    RDF.type,
                    ClassDefinition.class_class_uri,
                )
            )

        # Parent classes: is_a and mixins
        has_parent = False
        if cls.is_a:
            self.graph.add((cls_uri, RDFS.subClassOf, self._class_uri(cls.is_a)))
            has_parent = True
        for mixin in sorted(cls.mixins):
            parent = self._class_uri(mixin)
            if self.mixins_as_expressions:
                parent = self._some_values_from(self._metaslot_uri("mixins"), parent)
            else:
                has_parent = True
            self.graph.add((cls_uri, RDFS.subClassOf, parent))
        if not has_parent and self.add_root_classes:
            # If user selects add_root_classes, then all classes will be subclasses of LinkML:ClassDefinition
            if cls.mixin and self.mixins_as_expressions:
                self.graph.add((cls_uri, RDFS.subClassOf, self._mixin_grouping_class_uri()))
            else:
                self.graph.add((cls_uri, RDFS.subClassOf, URIRef(ClassDefinition.class_class_uri)))
        if self.has_profile(MetadataProfile.ols):
            # Add annotations for browser hints. See https://www.ebi.ac.uk/ols/docs/installation-guide
            if cls.is_a is None:
                if len(cls.mixins) == 0:
                    # Any class that is not a mixin and is a root serves as a potential entry point
                    self.graph.add(
                        (
                            self.graph.identifier,
                            URIRef("http://purl.obolibrary.org/obo/IAO_0000700"),
                            cls_uri,
                        )
                    )
        if cls.class_uri:
            # If a class_ur is assigned, and it is different from model class_uri, then
            # Add an assertion that links the two
            mapped_uri = sv.get_uri(cls, expand=True, native=not self.use_native_uris)
            if cls_uri != mapped_uri:
                p = OWL.equivalentClass if self.assert_equivalent_classes else SKOS.exactMatch
                self.graph.add((URIRef(cls_uri), p, URIRef(mapped_uri)))
        subject_expr = URIRef(cls_uri)
        if self.mixins_as_expressions and cls.mixin:
            subject_expr = self._some_values_from(self._metaslot_uri("mixins"), subject_expr)
        # type designator yield GCI rules
        # E.g. if C has a type designator slot t,
        # then create an axiom: (t some C) subClassOf C.
        type_designator = sv.get_type_designator_slot(cls.name)
        if type_designator:
            td_prop = self._prop_uri(type_designator.name)
            restr = self._some_values_from(td_prop, subject_expr)
            self.graph.add((restr, RDFS.subClassOf, subject_expr))
        # unique key constraints are mapped to OWL hasKey
        if cls.unique_keys:
            for uk in cls.unique_keys.values():
                uk_props = [self._prop_uri(slot) for slot in uk.unique_key_slots]
                uk_props_listnode = BNode()
                Collection(self.graph, uk_props_listnode, uk_props)
                self.graph.add((subject_expr, OWL.hasKey, uk_props_listnode))

        def condition_to_bnode(expr: AnonymousClassExpression) -> Optional[BNode]:
            # inner function: translate a LinkML class expression to an OWL class expression.
            ixn_listnode = self.transform_class_expression(expr, quantifier_predicate=OWL.someValuesFrom)
            if not ixn_listnode:
                return None
            if expr.is_a:
                ixn_listnode = self._intersection_of([ixn_listnode, self._class_uri(expr.is_a)])
            return ixn_listnode

        # rules yield OWL GCI subClassOf axioms
        for rule in cls.rules:
            pre_node = condition_to_bnode(rule.preconditions)
            if not pre_node:
                continue
            pre_node = self._intersection_of([pre_node, subject_expr])
            post_node = condition_to_bnode(rule.postconditions)
            if not post_node:
                continue
            self.graph.add((pre_node, RDFS.subClassOf, post_node))
        # classification rules yield OWL GCI subClassOf axioms
        for expr in cls.classification_rules:
            ixn_listnode = condition_to_bnode(expr)
            self.graph.add((ixn_listnode, RDFS.subClassOf, subject_expr))
        # Other axioms, including those from anonymous expressions
        superclass_expr = self.transform_class_expression(cls)
        if superclass_expr:
            ixn_listnodes = []
            if isinstance(superclass_expr, BNode):
                ixn_listnodes = list(self.graph.objects(superclass_expr, OWL.intersectionOf))
            if self.simplify and ixn_listnodes:
                # simplify
                if len(ixn_listnodes) > 1:
                    raise AssertionError
                ixn_listnode = ixn_listnodes[0]
                if not isinstance(ixn_listnode, BNode):
                    raise AssertionError
                for x in Collection(self.graph, ixn_listnode):
                    self.graph.add((subject_expr, RDFS.subClassOf, x))
                self._remove_list(ixn_listnode)
                self.graph.remove((superclass_expr, OWL.intersectionOf, ixn_listnodes[0]))
            else:
                self.graph.add((subject_expr, RDFS.subClassOf, superclass_expr))

    def get_own_slots(self, cls: Union[ClassDefinition, AnonymousClassExpression]) -> list[SlotDefinition]:
        """
        Get the slots that are defined on a class, excluding those that are inherited.

        :param cls:
        :return:
        """
        sv = self.schemaview
        if isinstance(cls, ClassDefinition):
            own_slots = (
                list(cls.slot_usage.values()) + list(cls.attributes.values()) + list(cls.slot_conditions.values())
            )
            for slot_name in cls.slots:
                # if slot_name not in cls.slot_usage:
                slot = sv.get_slot(slot_name)
                if slot:
                    own_slots.append(slot)
                else:
                    logger.warning(f"Unknown top-level slot {slot_name}")
        else:
            own_slots = []
        own_slots.extend(cls.slot_conditions.values())
        # merge slots with the same name
        slot_map = {}
        for slot in own_slots:
            if slot.name in slot_map:
                for k, v in slot.__dict__.items():
                    curr = slot_map[slot.name].get(k, None)
                    if v and not curr:
                        slot_map[slot.name][k] = v
            else:
                slot_map[slot.name] = copy(slot.__dict__)
        own_slots = [SlotDefinition(**v) for v in slot_map.values()]
        # sort by name
        own_slots.sort(key=lambda x: x.name)
        return own_slots

    def transform_class_expression(
        self,
        cls: Union[ClassDefinition, AnonymousClassExpression],
        quantifier_predicate: URIRef = OWL.allValuesFrom,
    ) -> BNode:
        """
        Transform a LinkML class expression into an OWL expression.

        If the class includes boolean expressions, then these
        are recursively transformed (each such inner expression is
        an anonymous expression)

        :param cls: LinkML class expression (anonymous if called recursively)
        :param quantifier_predicate:
        :return: blank node representing the OWL expression
        """
        if cls is None:
            cls = AnonymousClassExpression()
        graph = self.graph
        sv = self.schemaview
        own_slots = self.get_own_slots(cls)
        owl_exprs = []
        if cls.any_of:
            owl_exprs.append(self._union_of([self.transform_class_expression(x) for x in cls.any_of]))
        if cls.exactly_one_of:
            sub_exprs = [self.transform_class_expression(x) for x in cls.exactly_one_of]
            if isinstance(cls, ClassDefinition):
                cls_uri = self._class_uri(cls.name)
                listnode = BNode()
                Collection(graph, listnode, sub_exprs)
                graph.add((cls_uri, OWL.disjointUnionOf, listnode))
            else:
                sub_sub_exprs = []
                for i, x in enumerate(cls.exactly_one_of):
                    rest = cls.exactly_one_of[0:i] + cls.exactly_one_of[i + 1 :]
                    neg_expr = self._complement_of_union_of([self.transform_class_expression(nx) for nx in rest])
                    pos_expr = self._intersection_of([self.transform_class_expression(x), neg_expr])
                    sub_sub_exprs.append(pos_expr)
                owl_exprs.append(self._union_of(sub_sub_exprs))
                # owl_exprs.extend(sub_exprs)
        if cls.all_of:
            owl_exprs.append(self._intersection_of([self.transform_class_expression(x) for x in cls.all_of]))
        if cls.none_of:
            owl_exprs.append(self._complement_of_union_of([self.transform_class_expression(x) for x in cls.none_of]))
        for slot in own_slots:
            if slot.name:
                owltypes = self.slot_node_owltypes(sv.get_slot(slot.name), owning_class=cls)
            else:
                owltypes = self.slot_node_owltypes(slot, owning_class=cls)
            x = self.transform_class_slot_expression(cls, slot, slot, owltypes)
            if not x:
                range = sv.schema.default_range
                if range and OWL.Thing not in owltypes:
                    if range in sv.all_types():
                        x = self._type_uri(range)
                    elif range in sv.all_enums():
                        x = self._enum_uri(range)
                    elif range in sv.all_classes():
                        x = self._class_uri(range)
                    else:
                        raise ValueError(f"Unknown range {range}")
                        # x = self._class_uri(range)
                else:
                    x = OWL.Thing
            slot_uri = self._prop_uri(slot)
            if slot.name in sv.all_slots():
                top_slot = sv.get_slot(slot.name)
            else:
                top_slot = slot
            avf = BNode()
            graph.add((avf, RDF.type, OWL.Restriction))
            graph.add((avf, quantifier_predicate, x))
            graph.add((avf, OWL.onProperty, slot_uri))
            owl_exprs.append(avf)
            if isinstance(cls, AnonymousClassExpression):
                # cardinality constraints only belong at the top level
                continue
            min_card_expr = BNode()
            owl_exprs.append(min_card_expr)
            min_card = 1 if slot.required or top_slot.required else 0
            graph.add((min_card_expr, RDF.type, OWL.Restriction))
            graph.add((min_card_expr, OWL.minCardinality, Literal(min_card)))
            graph.add((min_card_expr, OWL.onProperty, slot_uri))
            if not slot.multivalued and not top_slot.multivalued:
                max_card_expr = BNode()
                owl_exprs.append(max_card_expr)
                graph.add((max_card_expr, RDF.type, OWL.Restriction))
                graph.add((max_card_expr, OWL.maxCardinality, Literal(1)))
                graph.add((max_card_expr, OWL.onProperty, slot_uri))
            if slot.has_member:
                has_member_expr = self.transform_class_slot_expression(cls, slot.has_member, slot)
                if has_member_expr:
                    owl_exprs.append(self._some_values_from(slot_uri, has_member_expr))
        return self._intersection_of(owl_exprs)

    def slot_node_owltypes(
        self,
        slot: Union[SlotDefinition, AnonymousSlotExpression],
        owning_class: Optional[Union[ClassDefinition, AnonymousClassExpression]] = None,
    ) -> set[URIRef]:
        """
        Determine the OWL types of a named slot or slot expression

        The OWL type is either OWL.Thing or RDFS.Datatype

        :param slot:
        :param owning_class:
        :return:
        """
        sv = self.schemaview
        node_types = set()
        if isinstance(slot, SlotDefinition):
            slot_range = slot.range
            if isinstance(owning_class, ClassDefinition):
                slot_range = sv.induced_slot(slot.name, owning_class.name).range
            if slot_range in sv.all_classes():
                range_class = sv.get_class(slot_range)
                if not (range_class and range_class.class_uri == "linkml:Any"):
                    node_types.add(OWL.Thing)
            if slot.range in sv.all_types():
                node_types.add(RDFS.Datatype)
        for k in ["any_of", "all_of", "exactly_one_of", "none_of"]:
            subslot = getattr(slot, k, None)
            if subslot:
                node_types.update(self.slot_node_owltypes(subslot, owning_class=owning_class))
        return node_types

    def transform_class_slot_expression(
        self,
        cls: Optional[Union[ClassDefinition, AnonymousClassExpression]],
        slot: Union[SlotDefinition, AnonymousSlotExpression],
        main_slot: SlotDefinition = None,
        owl_types: set[OWL_TYPE] = None,
    ) -> Optional[Union[BNode, URIRef]]:
        """
        Take a ClassExpression and SlotExpression combination and transform to a node.

        :param cls:
        :param slot:
        :param main_slot:
        :param owl_types:
        :return:
        """
        sv = self.schemaview
        if main_slot is None:
            if not isinstance(slot, SlotDefinition):
                raise ValueError(f"Must pass main slot for {slot}")
            main_slot = slot

        owl_exprs = []

        if slot.all_members:
            owl_exprs.append(self.transform_class_slot_expression(cls, slot.all_members, main_slot, owl_types))

        if slot.any_of:
            owl_exprs.append(
                self._union_of(
                    [self.transform_class_slot_expression(cls, x, main_slot, owl_types) for x in slot.any_of]
                )
            )
        if slot.all_of:
            owl_exprs.append(
                self._intersection_of(
                    [self.transform_class_slot_expression(cls, x, main_slot, owl_types) for x in slot.all_of]
                )
            )
        if slot.none_of:
            owl_exprs.append(
                self._complement_of_union_of(
                    [self.transform_class_slot_expression(cls, x, main_slot, owl_types) for x in slot.none_of]
                )
            )
        if slot.exactly_one_of:
            disj_exprs = []
            for i, operand in enumerate(slot.exactly_one_of):
                rest = slot.exactly_one_of[0:i] + slot.exactly_one_of[i + 1 :]
                neg_expr = self._complement_of_union_of(
                    [self.transform_class_slot_expression(cls, x, main_slot, owl_types) for x in rest],
                    owl_types=owl_types,
                )
                pos_expr = self._intersection_of(
                    [self.transform_class_slot_expression(cls, operand, main_slot, owl_types), neg_expr],
                    owl_types=owl_types,
                )
                disj_exprs.append(pos_expr)
            owl_exprs.append(self._union_of(disj_exprs, owl_types=owl_types))
        range = slot.range
        # if not range and not owl_exprs:
        #    range = sv.schema.default_range
        this_owl_types = set()
        if range:
            if range in sv.all_types(imports=True):
                self.slot_is_literal_map[main_slot.name].add(True)
                this_owl_types.add(RDFS.Literal)
                typ = sv.get_type(range)
                if self.type_objects:
                    # TODO
                    owl_exprs.append(self._type_uri(typ.name))
                else:
                    owl_exprs.append(self._type_uri(typ.name))
            elif range in sv.all_enums(imports=True):
                # TODO: enums fill this in
                owl_exprs.append(self._enum_uri(EnumDefinitionName(range)))
            elif range in sv.all_classes(imports=True):
                this_owl_types.add(OWL.Thing)
                self.slot_is_literal_map[main_slot.name].add(False)
                owl_exprs.append(self._class_uri(ClassDefinitionName(range)))
            else:
                raise ValueError(f"Unknown range {range}")
        is_literal = None
        if owl_types:
            is_literal = RDFS.Datatype in owl_types
        constraints_exprs, constraints_owltypes = self.add_constraints(slot, is_literal=is_literal)
        this_owl_types.update(constraints_owltypes)
        owl_exprs.extend(constraints_exprs)
        this_expr = self._intersection_of(owl_exprs, owl_types=this_owl_types)
        self.node_owltypes[this_expr].update(self._get_owltypes(this_owl_types, owl_exprs))
        return this_expr

    def add_constraints(
        self,
        element: Union[SlotDefinition, AnonymousSlotExpression, TypeDefinition, AnonymousTypeExpression],
        is_literal: Optional[bool] = None,
    ) -> tuple[list[BNode], set[OWL_TYPE]]:
        owl_types = set()
        owl_exprs = []
        graph = self.graph
        constraints = {
            XSD.minInclusive: element.minimum_value,
            XSD.maxInclusive: element.maximum_value,
            XSD.pattern: element.pattern,  # TODO: map between ECMAScript and XSD regular expressions
        }
        if element.equals_number is not None:
            constraints[XSD.minInclusive] = element.equals_number
            constraints[XSD.maxInclusive] = element.equals_number
        if element.equals_string is not None:
            equals_string = element.equals_string
            if is_literal is None:
                logger.warning(f"ignoring equals_string={equals_string} as unable to tell if literal")
            elif is_literal:
                constraints[XSD.pattern] = equals_string
            else:
                eq_uri = URIRef(self.schemaview.expand_curie(equals_string))
                owl_exprs.append(eq_uri)
        if element.equals_string_in:
            equals_string_in = element.equals_string_in
            literals = [Literal(s) for s in equals_string_in]
            one_of_expr = self._boolean_expression(literals, OWL.oneOf, owl_types={RDFS.Literal})
            owl_exprs.append(one_of_expr)
            owl_types.add(RDFS.Literal)
        for constraint_prop, constraint_val in constraints.items():
            if is_literal is not None and not is_literal:
                # In LinkML, it is permissible to have a literal constraints on slots that refer to
                # other objects. E.g. a pattern on a in_organization slot which refers to an Organization
                # will be applied to the id of that Organization.
                # To support this in OWL we would need to change this to a complex expression - for
                # now we will skip this.
                # See: https://github.com/linkml/linkml/issues/1841
                continue
            if constraint_val is not None:
                owl_types.add(RDFS.Literal)
                dr = BNode()
                graph.add((dr, RDF.type, RDFS.Datatype))
                if isinstance(constraint_val, float):
                    graph.add((dr, OWL.onDatatype, XSD.float))
                elif isinstance(constraint_val, int):
                    graph.add((dr, OWL.onDatatype, XSD.integer))
                else:
                    graph.add((dr, OWL.onDatatype, XSD.string))
                listnode = BNode()
                x = BNode()
                Collection(graph, listnode, [x])
                graph.add((dr, OWL.withRestrictions, listnode))
                graph.add((x, constraint_prop, Literal(constraint_val)))
                owl_exprs.append(dr)
        return owl_exprs, owl_types

    def add_slot(self, slot: SlotDefinition, attribute=False) -> None:
        # determine if this is a slot that has been induced by slot_usage; if so
        # the meaning of the slot is context-specific and should not be used for
        # global properties

        slot_uri = self._prop_uri(slot)

        # Slots may be modeled as Object or Datatype Properties
        # if type_objects is True, then ALL slots are ObjectProperties
        self.graph.add((slot_uri, RDF.type, self.slot_owl_type(slot)))
        if self.metaclasses:
            # add metaclass which this property instantiates -- induces punning
            self.graph.add(
                (
                    slot_uri,
                    RDF.type,
                    SlotDefinition.class_class_uri,
                )
            )

        if attribute:
            n = 0
            for c in self.schemaview.all_classes().values():
                for a in c.attributes.values():
                    att_uri = self.schemaview.get_uri(a, native=False, expand=True)
                    if slot_uri == URIRef(att_uri):
                        n += 1
            if n > 1:
                logger.warning(f"Ambiguous attribute: {slot.name} {slot_uri}")
                return

        self.add_metadata(slot, slot_uri)

        if attribute:
            return

        range_expr = self.transform_class_slot_expression(None, slot)
        if range_expr:
            self.graph.add((slot_uri, RDFS.range, range_expr))
        if slot.domain:
            self.graph.add((slot_uri, RDFS.domain, self._class_uri(slot.domain)))
        if slot.inverse:
            self.graph.add((slot_uri, OWL.inverseOf, self._prop_uri(slot.inverse)))
        characteristics = {
            "symmetric": OWL.SymmetricProperty,
            "asymmetric": OWL.AsymmetricProperty,
            "transitive": OWL.TransitiveProperty,
            "reflexive": OWL.ReflexiveProperty,
            "irreflexive": OWL.IrreflexiveProperty,
        }
        for metaslot, uri in characteristics.items():
            if getattr(slot, metaslot, False):
                self.graph.add((slot_uri, RDF.type, uri))

        if slot.is_a:
            self.graph.add((slot_uri, RDFS.subPropertyOf, self._prop_uri(slot.is_a)))
        for mixin in slot.mixins:
            self.graph.add((slot_uri, RDFS.subPropertyOf, self._prop_uri(mixin)))

    def add_type(self, typ: TypeDefinition) -> None:
        type_uri = self._type_uri(typ.name)
        if typ.from_schema == "https://w3id.org/linkml/types":
            return

        if self.metaclasses:
            self.graph.add(
                (
                    type_uri,
                    RDF.type,
                    URIRef(TypeDefinition.class_class_uri),
                )
            )
        # self._add_element_properties(type_uri, typ)
        if self.type_objects:
            self.graph.add((type_uri, RDF.type, OWL.Class))
            if typ.typeof:
                self.graph.add((type_uri, RDFS.subClassOf, self._type_uri(typ.typeof)))
            else:
                if not self.top_value_uri:
                    self.top_value_uri = self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME]["topValue"]
                    self.graph.add((self.top_value_uri, RDF.type, OWL.DatatypeProperty))
                    self.graph.add((self.top_value_uri, RDFS.label, Literal("value")))
                restr = BNode()
                self.graph.add((restr, RDF.type, OWL.Restriction))
                self.graph.add((restr, OWL.qualifiedCardinality, Literal(1)))
                self.graph.add((restr, OWL.onProperty, self.top_value_uri))
                self.graph.add((restr, OWL.onDataRange, self._type_uri(typ.name)))
                self.graph.add((type_uri, RDFS.subClassOf, restr))
        else:
            self.graph.add((type_uri, RDF.type, RDFS.Datatype))
            eq_conjunctions = []
            if typ.typeof and type_uri != self._type_uri(typ.typeof):
                # self.graph.add((type_uri, OWL.equivalentClass, self._type_uri(typ.typeof)))
                eq_conjunctions.append(self._type_uri(typ.typeof))
                # self.graph.add((type_uri, RDFS.subClassOf, self._type_uri(typ.typeof)))
            if typ.uri and type_uri != URIRef(self.schemaview.expand_curie(typ.uri)):
                eq_conjunctions.append(URIRef(self.schemaview.expand_curie(typ.uri)))
                # self.graph.add(
                #    (type_uri, OWL.equivalentClass, URIRef(self.schemaview.expand_curie(typ.uri)))
                # )
            constraints_exprs, _ = self.add_constraints(typ, is_literal=True)
            eq_conjunctions.extend(constraints_exprs)
            ixn = self._intersection_of(eq_conjunctions)
            if ixn:
                self.graph.add((type_uri, OWL.equivalentClass, ixn))

    def _get_metatype(
        self, element: Union[Definition, PermissibleValue], default_value: Optional[Union[str, URIRef]] = None
    ) -> Optional[URIRef]:
        impls = []
        if isinstance(element, Definition):
            impls.extend(element.implements)
        if isinstance(element, PermissibleValue):
            if "implements" in element.annotations:
                ann = element.annotations["implements"]
                v = ann.value
                if not isinstance(v, list):
                    v = [v]
                impls.extend(v)
            impls.extend(element.implements)
        for impl in impls:
            if impl.startswith("owl:"):
                return OWL[impl.split(":")[1]]
            if impl.startswith("rdfs:"):
                return RDFS[impl.split(":")[1]]
        if isinstance(default_value, str):
            if default_value.startswith("owl:"):
                return OWL[default_value.split(":")[1]]
            if default_value.startswith("rdfs:"):
                return RDFS[default_value.split(":")[1]]
            return URIRef(default_value)
        return default_value

    def add_enum(self, e: EnumDefinition) -> None:
        g = self.graph
        enum_uri = self._enum_uri(e.name)
        g.add((enum_uri, RDF.type, OWL.Class))

        self.add_metadata(e, enum_uri)
        has_parent = False
        if e.is_a:
            self.graph.add((enum_uri, RDFS.subClassOf, self._enum_uri(e.is_a)))
            has_parent = True
        for mixin in sorted(e.mixins):
            parent = self._enum_uri(mixin)
            if self.mixins_as_expressions:
                parent = self._some_values_from(self._metaslot_uri("mixins"), parent)
            else:
                has_parent = True
            self.graph.add((enum_uri, RDFS.subClassOf, parent))
        if not has_parent and self.add_root_classes:
            self.graph.add((enum_uri, RDFS.subClassOf, URIRef(EnumDefinition.class_class_uri)))
        if self.metaclasses:
            g.add(
                (
                    enum_uri,
                    RDF.type,
                    URIRef(EnumDefinition.class_class_uri),
                )
            )
        pv_uris = []
        owl_types = []
        enum_owl_type = self._get_metatype(e, self.default_permissible_value_type)

        for pv in e.permissible_values.values():
            pv_owl_type = self._get_metatype(pv, enum_owl_type)
            owl_types.append(pv_owl_type)
            if pv_owl_type == RDFS.Literal:
                pv_node = Literal(pv.text)
                if pv.meaning:
                    logger.warning(f"Meaning on literal {pv.text} in {e.name} is ignored")
            else:
                pv_node = self._permissible_value_uri(pv, enum_uri, e)
            pv_uris.append(pv_node)
            g.add(
                (
                    enum_uri,
                    self.metamodel.namespaces[METAMODEL_NAMESPACE_NAME]["permissible_values"],
                    pv_node,
                )
            )
            if not isinstance(pv_node, Literal):
                self.add_metadata(pv, pv_node)
                g.add((pv_node, RDF.type, pv_owl_type))
                g.add((pv_node, RDFS.label, Literal(pv.text)))
                # TODO: make this configurable
                # self._add_element_properties(pv_uri, pv)
                if self.metaclasses:
                    g.add((pv_node, RDF.type, enum_uri))
                has_parent = False
                if pv.is_a:
                    self.graph.add((pv_node, RDFS.subClassOf, self._permissible_value_uri(pv.is_a, enum_uri, e)))
                    has_parent = True
                for mixin in sorted(pv.mixins):
                    parent = self._permissible_value_uri(mixin, enum_uri, e)
                    if self.mixins_as_expressions:
                        parent = self._some_values_from(self._metaslot_uri("mixins"), parent)
                    else:
                        has_parent = True
                    self.graph.add((enum_uri, RDFS.subClassOf, parent))
                if not has_parent and self.add_root_classes:
                    self.graph.add((pv_node, RDFS.subClassOf, URIRef(PermissibleValue.class_class_uri)))
        if all([pv is not None for pv in pv_uris]):
            # every single PV in the enum is not-null
            all_is_class = all([owl_type == OWL.Class for owl_type in owl_types])
            all_is_individual = all([owl_type == OWL.NamedIndividual for owl_type in owl_types])
            all_is_literal = all([owl_type == RDFS.Literal for owl_type in owl_types])
            sub_pred = DCTERMS.isPartOf
            combo_pred = None
            if all_is_class or all_is_individual or all_is_literal:
                if all_is_class:
                    combo_pred = OWL.unionOf
                    # self._union_of(pv_uris, node=enum_uri)
                    sub_pred = RDFS.subClassOf
                elif all_is_individual:
                    combo_pred = OWL.oneOf
                    # self._object_one_of(pv_uris, node=enum_uri)
                    sub_pred = RDF.type
                elif all_is_literal:
                    combo_pred = OWL.oneOf
                    # self._object_one_of(pv_uris, node=enum_uri)
                    sub_pred = RDF.type
                if combo_pred:
                    self._boolean_expression(pv_uris, combo_pred, enum_uri, owl_types=set(owl_types))
            for pv_node in pv_uris:
                # this would normally be entailed, but we assert here so it is visible
                # without reasoning
                if not isinstance(pv_node, Literal):
                    g.add((pv_node, sub_pred, enum_uri))

    def _add_rule(self, subject: Union[URIRef, BNode], rule: ClassRule, cls: ClassDefinition):
        if not self.use_swrl:
            return
        logger.warning("SWRL support is experimental and incomplete")
        head = []
        body = []
        for pre in rule.preconditions:
            head.extend(self._add_rule_condition(subject, pre, cls))
        for post in rule.postconditions:
            body.extend(self._add_rule_condition(subject, post, cls))
        self._swrl_rule(subject, body, head)

    def _add_rule_condition(
        self,
        subject: Union[URIRef, BNode],
        condition: AnonymousClassExpression,
        cls: ClassDefinition,
    ) -> list[BNode]:
        for slot_name, expr in condition.slot_conditions.items():
            var = self._swrl_var(slot_name)
            if expr.maximum_value is not None:
                self.graph.add((var, SWRLB.lessThanOrEqual, Literal(expr.maximum_value)))

    def has_profile(self, profile: MetadataProfile, default=False) -> bool:
        """
        Determine if a metadata profile is active.

        :param profile: profile to check
        :param default: True if the configured profiles include the specified profile
        :return:
        """
        if default and not self.metadata_profile and not self.metadata_profiles:
            return True
        return profile in self.metadata_profiles or profile == self.metadata_profile

    def _get_owltypes(self, current: set[OWL_TYPE], exprs: list[Union[BNode, URIRef]]) -> set[OWL_TYPE]:
        """
        Gets the OWL types of specified expressions plus current owl types.

        :param current:
        :param exprs:
        :return:
        """
        owltypes = set()
        for x in exprs:
            x_owltypes = self.node_owltypes.get(x, None)
            if x_owltypes:
                owltypes.update(x_owltypes)
        owltypes.update(current)
        if len(owltypes) > 1:
            logger.warning(f"Multiple owl types {owltypes}")
            # if self.target_profile == OWLProfile.dl:
        return owltypes

    def _remove_list(self, listnode: BNode) -> None:
        graph = self.graph
        triples = list(graph.triples((listnode, None, None)))
        while triples:
            t = triples.pop(0)
            subj, pred, obj = t
            if pred not in [RDF.first, RDF.rest]:
                continue
            graph.remove(t)
            if isinstance(obj, BNode):
                triples.extend(graph.triples((obj, None, None)))

    def _some_values_from(self, property: URIRef, filler: Union[URIRef, BNode]) -> BNode:
        if not property:
            raise ValueError(f"Property is required, filler: {filler}")
        if not filler:
            raise ValueError(f"Filler is required, property: {property}")
        node = BNode()
        self.graph.add((node, RDF.type, OWL.Restriction))
        self.graph.add((node, OWL.onProperty, property))
        self.graph.add((node, OWL.someValuesFrom, filler))
        return node

    def _has_value(self, property: URIRef, filler: Union[URIRef, BNode, Literal]) -> BNode:
        node = BNode()
        self.graph.add((node, RDF.type, OWL.Restriction))
        self.graph.add((node, OWL.onProperty, property))
        self.graph.add((node, OWL.hasValue, filler))
        return node

    @staticmethod
    def _swrl_var(var: str) -> URIRef:
        return URIRef(var)

    def _swrl_class_atom(self, cls_ref: Union[BNode, URIRef], var: str) -> BNode:
        node = BNode()
        self.graph.add((node, RDF.type, SWRL.ClassAtom))
        self.graph.add((node, SWRL.classPredicate, cls_ref))
        self.graph.add((node, SWRL.argument1, self._swrl_var(var)))
        return node

    def _swrl_abox_atom(self, pred_ref: Union[BNode, URIRef], arg1: str, arg2: str) -> BNode:
        node = BNode()
        self.graph.add((node, RDF.type, SWRL.IndividualPropertyAtom))
        self.graph.add((node, SWRL.classPredicate, pred_ref))
        self.graph.add((node, SWRL.argument1, self._swrl_var(arg1)))
        self.graph.add((node, SWRL.argument2, self._swrl_var(arg2)))
        return node

    def _swrl_rule(self, body, head) -> BNode:
        node = BNode()
        self.graph.add((node, RDF.type, SWRL.Imp))
        self.graph.add((node, SWRL.body, body))
        self.graph.add((node, SWRL.head, head))
        return node

    @staticmethod
    def _metaslot_uri(name: str) -> URIRef:
        return URIRef("https://w3id.org/linkml/" + name)

    def _complement_of_union_of(
        self, exprs: list[Union[BNode, URIRef]], owl_types: set[OWL_TYPE] = None, **kwargs
    ) -> Optional[Union[BNode, URIRef]]:
        if not exprs:
            raise ValueError("Must pass at least one")
        neg_expr = BNode()
        if not owl_types:
            owl_types = self._get_owltypes(set(), exprs)
        complement_predicate = OWL.complementOf
        if len(owl_types) == 1:
            if RDFS.Literal in owl_types:
                self.graph.add((neg_expr, RDF.type, RDFS.Datatype))
                complement_predicate = OWL.datatypeComplementOf
        self.graph.add((neg_expr, complement_predicate, self._union_of(exprs)), **kwargs)

        return neg_expr

    def _intersection_of(self, exprs: list[Union[BNode, URIRef]], **kwargs) -> Optional[Union[BNode, URIRef]]:
        return self._boolean_expression(exprs, OWL.intersectionOf, **kwargs)

    def _union_of(self, exprs: list[Union[BNode, URIRef]], **kwargs) -> Optional[Union[BNode, URIRef]]:
        return self._boolean_expression(exprs, OWL.unionOf, **kwargs)

    def _object_one_of(self, exprs: list[Union[BNode, URIRef]], **kwargs) -> Optional[Union[BNode, URIRef]]:
        return self._boolean_expression(exprs, OWL.oneOf, **kwargs)

    def _exactly_one_of(self, exprs: list[Union[BNode, URIRef]]) -> Optional[Union[BNode, URIRef]]:
        if not exprs:
            raise ValueError("Must pass at least one")
        if len(exprs) == 1:
            return exprs[0]
        sub_exprs = []
        for i, x in enumerate(exprs):
            rest = exprs[0:i] + exprs[i + 1 :]
            neg_expr = self._complement_of_union_of(rest)
            sub_exprs.append(self._intersection_of([x, neg_expr]))
        return self._union_of(sub_exprs)

    def _datatype_restriction(self, datatype: URIRef, facets: list[Union[BNode, URIRef]]) -> BNode:
        node = BNode()
        graph = self.graph
        graph.add((node, RDF.type, RDFS.Datatype))
        graph.add((node, OWL.onDatatype, datatype))
        listnode = BNode()
        Collection(graph, listnode, facets)
        graph.add((node, OWL.withRestrictions, listnode))
        return node

    def _facet(self, typ: URIRef, val: Union[Literal, Any]):
        if not isinstance(val, Literal):
            val = Literal(val)
        node = BNode()
        self.graph.add((node, typ, val))
        return node

    def _boolean_expression(
        self,
        exprs: list[Union[BNode, URIRef, Literal]],
        predicate: URIRef,
        node: Optional[URIRef] = None,
        owl_types: set[OWL_TYPE] = None,
    ) -> Optional[Union[BNode, URIRef]]:
        graph = self.graph
        if [x for x in exprs if x is None]:
            logger.warning(f"Null expr in: {exprs} for {predicate} {node}")
            exprs = [x for x in exprs if x is not None]
        if len(exprs) == 0:
            return None
        elif len(exprs) == 1:
            return exprs[0]
        else:
            if node is None:
                node = BNode()
            listnode = BNode()
            Collection(graph, listnode, exprs)
            graph.add((node, predicate, listnode))
            if owl_types is None:
                owl_types = set()
            owl_types = owl_types.union(self._get_owltypes(set(), exprs))
            if len(owl_types) == 1:
                if RDFS.Literal in owl_types:
                    graph.add((node, RDF.type, RDFS.Datatype))
            return node

    def _range_is_datatype(self, slot: SlotDefinition) -> bool:
        if self.type_objects:
            return False
        else:
            return slot.range in self.schema.types

    def _range_uri(self, slot: SlotDefinition) -> URIRef:
        if slot.range in self.schema.types:
            typ = self.schema.types[slot.range]
            if self.type_objects:
                return self._type_uri(typ.name)
            else:
                return self.namespaces.uri_for(typ.uri)
        elif slot.range in self.schema.enums:
            # TODO: enums fill this in
            return self._enum_uri(EnumDefinitionName(slot.range))
        else:
            return self._class_uri(ClassDefinitionName(slot.range))

    @staticmethod
    def _mixin_grouping_class_uri():
        return URIRef(ClassDefinition.class_class_uri + "#Mixin")

    def _class_uri(self, cn: Union[str, ClassDefinitionName]) -> URIRef:
        c = self.schemaview.get_class(cn)
        return URIRef(self.schemaview.get_uri(c, expand=True, native=self.use_native_uris))

    def _enum_uri(self, en: Union[str, EnumDefinitionName]) -> URIRef:
        sv = self.schemaview
        e = sv.get_enum(en, strict=True)
        uri = e.enum_uri
        if not uri:
            uri = f"{sv.schema.default_prefix}:{camelcase(en)}"
        # TODO: allow control over URIs
        return URIRef(self.schemaview.expand_curie(uri))

    def _prop_uri(self, pn: Union[SlotDefinition, SlotDefinitionName]) -> URIRef:
        if isinstance(pn, SlotDefinition):
            p = pn
        else:
            p = self.schemaview.get_slot(pn, attributes=True)
        if not p:
            raise ValueError(f"No such slot or attribute: {pn}")
        try:
            return URIRef(self.schemaview.get_uri(p, expand=True, native=self.use_native_uris))
        except (KeyError, ValueError):
            # TODO: fix this upstream in schemaview
            default_prefix = self.schemaview.schema.default_prefix or ""
            return URIRef(self.schemaview.expand_curie(f"{default_prefix}:{underscore(p.name)}"))

    def _schema_uri(self, scn: Union[str, SchemaDefinitionName]) -> URIRef:
        if ":" in scn:
            return URIRef(self.schemaview.expand_curie(scn))
        else:
            default_prefix = self.schemaview.schema.default_prefix or ""
            return URIRef(self.schemaview.expand_curie(f"{default_prefix}:{scn}"))

    def _type_uri(self, tn: TypeDefinitionName, native: bool = None) -> URIRef:
        if native is None:
            # never use native unless type shadowing with objects is enabled
            native = self.type_objects
        if native:
            # UGLY HACK: Currently schemaview does not camelcase types
            e = self.schemaview.get_element(tn, imports=True)
            if e.from_schema is not None:
                schema = next(sc for sc in self.schemaview.schema_map.values() if sc.id == e.from_schema)
                pfx = schema.default_prefix
                if pfx == "linkml":
                    return URIRef(self.schemaview.expand_curie(f"{pfx}:{camelcase(tn)}"))
        t = self.schemaview.get_type(tn)
        expanded = self.schemaview.get_uri(t, expand=True, native=native)
        if expanded.startswith("xsd:"):
            # TODO: fix upstream in schemaview; default_curi_maps is different on windows
            return XSD[expanded[4:]]
        return URIRef(expanded)

    def _permissible_value_uri(
        self, pv: Union[str, PermissibleValue], enum_uri: str, enum_def: EnumDefinition = None
    ) -> URIRef:
        if isinstance(pv, str):
            pv_name = pv
            if enum_def is None:
                raise ValueError(f"Cannot find permissible value: {pv}, no enum definition provided")
            pvs = [pv for k, pv in enum_def.permissible_values.items() if k == pv_name]
            if len(pvs) != 1:
                raise ValueError(f"Cannot find permissible value: {pv_name}, got: {pvs}")
            pv = pvs[0]
        if pv.meaning:
            return URIRef(self.schemaview.expand_curie(pv.meaning))
        else:
            from urllib.parse import quote

            encoded_text = quote(pv.text.strip(), safe="", encoding="utf-8")
            return URIRef(enum_uri + self.enum_iri_separator + encoded_text)

    def slot_owl_type(self, slot: SlotDefinition) -> URIRef:
        sv = self.schemaview
        if slot.implements:
            for t in ["owl:AnnotationProperty", "owl:ObjectProperty", "owl:DatatypeProperty"]:
                if t in slot.implements:
                    return OWL[t.replace("owl:", "")]
        if slot.range is None:
            range = self.schemaview.schema.default_range
        else:
            range = slot.range
        if self.type_objects:
            return OWL.ObjectProperty
        is_literal_vals = self.slot_is_literal_map[slot.name]
        if len(is_literal_vals) > 1:
            logger.warning(f"Ambiguous type for: {slot.name}")
        if range is None:
            if not is_literal_vals:
                logger.warning(f"Guessing type for {slot.name}")
                return OWL.ObjectProperty
            if (list(is_literal_vals))[0]:
                return OWL.DatatypeProperty
            else:
                return OWL.ObjectProperty
        elif range in sv.all_classes():
            return OWL.ObjectProperty
        elif range in sv.all_enums():
            return OWL.ObjectProperty
        elif range in sv.all_types():
            return OWL.DatatypeProperty
        else:
            raise Exception(f"Unknown range: {slot.range}")


@shared_arguments(OwlSchemaGenerator)
@click.command(name="owl")
@click.option("-o", "--output", help="Output file name")
@click.option(
    "--metadata-profile",
    default=MetadataProfile.linkml.value,
    show_default=True,
    type=click.Choice(MetadataProfile.list()),
    help="What kind of metadata profile to use for annotations on generated OWL objects",
)
@click.option(
    "--type-objects/--no-type-objects",
    default=False,
    show_default=True,
    help="If true, will model linkml types as objects, not literals",
)
@click.option(
    "--metaclasses/--no-metaclasses",
    default=False,
    show_default=True,
    help="If true, include linkml metamodel classes as metaclasses. Note this introduces punning in OWL-DL",
)
@click.option(
    "--add-root-classes/--no-add-root-classes",
    default=False,
    show_default=True,
    help="If true, include linkml metamodel classes as superclasses.",
)
@click.option(
    "--add-ols-annotations/--no-add-ols-annotations",
    default=True,
    show_default=True,
    help="If true, auto-include annotations from https://www.ebi.ac.uk/ols/docs/installation-guide",
)
@click.option(
    "--ontology-uri-suffix",
    default=".owl.ttl",
    show_default=True,
    help="Suffix to append to schema id to generate OWL Ontology IRI",
)
@click.option(
    "--assert-equivalent-classes/--no-assert-equivalent-classes",
    default=False,
    show_default=True,
    help="If true, add owl:equivalentClass between a class and a class_uri",
)
@click.option(
    "--mixins-as-expressions/--no-mixins-as-expressions",
    default=False,
    show_default=True,
    help="If true, then mixins are represented as existential expressions",
)
@click.option(
    "--use-native-uris/--no-use-native-uris",
    default=True,
    show_default=True,
    help="Use model URIs rather than class/slot URIs",
)
@click.option(
    "--default-permissible-value-type",
    default=str(OWL.Class),
    show_default=True,
    help="Default OWL type for permissible values",
)
@click.option(
    "--enum-iri-separator",
    default="#",
    is_flag=False,
    show_default=True,
    help="IRI separator for enums.",
)
@click.version_option(__version__, "-V", "--version")
def cli(yamlfile, metadata_profile: str, **kwargs):
    """Generate an OWL representation of a LinkML model

    Generate OWL using default parameters:

        gen-owl my_schema.yaml

    Note that in previous versions of this generator, the default was to use type objects and
    to include metaclasses. To restore this behavior:

        gen-owl --metaclasses --type-objects my_schema.yaml

    For more info, see: https://linkml.io/linkml/generators/owl
    """
    if metadata_profile is not None:
        metadata_profiles = [MetadataProfile(metadata_profile)]
    else:
        metadata_profiles = [MetadataProfile.linkml]
    gen = OwlSchemaGenerator(yamlfile, metadata_profiles=metadata_profiles, **kwargs)
    print(gen.serialize(**kwargs))


if __name__ == "__main__":
    cli()



import json
import os
from functools import lru_cache
from typing import Optional

import jsonschema
from jsonschema.protocols import Validator
from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import SchemaDefinition

from linkml.generators import JsonSchemaGenerator, PydanticGenerator
from linkml.utils.datautils import infer_root_class


class ValidationContext:
    """Provides state that may be shared between validation plugins"""

    def __init__(self, schema: SchemaDefinition, target_class: Optional[str] = None) -> None:
        # Since SchemaDefinition is not hashable, to make caching simpler we store the schema
        # in a "private" property and assume it never changes.
        self._schema = schema
        self._schema_view = SchemaView(self._schema)
        self._target_class = self._get_target_class(target_class)

    @property
    def schema_view(self):
        return self._schema_view

    @property
    def target_class(self):
        return self._target_class

    @lru_cache
    def json_schema_validator(
        self,
        *,
        closed: bool,
        include_range_class_descendants: bool,
        path_override: Optional[os.PathLike] = None,
    ) -> Validator:
        if path_override:
            with open(path_override) as json_schema_file:
                json_schema = json.load(json_schema_file)
        else:
            not_closed = not closed
            jsonschema_gen = JsonSchemaGenerator(
                schema=self._schema,
                mergeimports=True,
                top_class=self._target_class,
                not_closed=not_closed,
                include_range_class_descendants=include_range_class_descendants,
            )
            json_schema = jsonschema_gen.generate()

        validator_cls = jsonschema.validators.validator_for(json_schema, default=jsonschema.Draft7Validator)
        return validator_cls(json_schema, format_checker=validator_cls.FORMAT_CHECKER)

    def pydantic_model(self, *, closed: bool):
        module = self._pydantic_module(closed=closed)
        return module.__dict__[self._target_class]

    @lru_cache
    def _pydantic_module(self, *, closed: bool):
        return PydanticGenerator(
            self._schema,
            extra_fields="forbid" if closed else "ignore" if closed is None else "allow",
        ).compile_module()

    def _get_target_class(self, target_class: Optional[str] = None) -> str:
        if target_class is None:
            return infer_root_class(self._schema_view)
        else:
            # strict=True raises ValueError if class is not found in schema
            class_def = self._schema_view.get_class(target_class, strict=True)
            return class_def.name



"""
Classes to inject in generated pydantic models
"""

LinkMLMeta = """
class LinkMLMeta(RootModel):
    root: dict[str, Any] = {}
    model_config = ConfigDict(frozen=True)

    def __getattr__(self, key:str):
        return getattr(self.root, key)

    def __getitem__(self, key:str):
        return self.root[key]

    def __setitem__(self, key:str, value):
        self.root[key] = value

    def __contains__(self, key:str) -> bool:
        return key in self.root

"""



import logging
import os
from dataclasses import dataclass
from enum import Enum
from pathlib import PurePosixPath
from types import ModuleType
from typing import Optional

import click
from jinja2 import Environment, PackageLoader
from linkml_runtime.linkml_model.meta import TypeDefinition
from linkml_runtime.utils.compile_python import compile_python
from linkml_runtime.utils.formatutils import camelcase
from linkml_runtime.utils.schemaview import SchemaView

from linkml._version import __version__
from linkml.generators.oocodegen import OOCodeGenerator

from .class_handler_base import ClassHandlerBase
from .enum_handler_base import EnumHandlerBase
from .render_adapters import DataframeDocument
from .render_adapters.dataframe_class import DataframeClass
from .slot_handler_base import SlotHandlerBase

logger = logging.getLogger(__name__)


# keys are template_path
TYPEMAP = {
    "panderagen_class_based": {
        "xsd:string": "str",
        "xsd:integer": "int",
        "xsd:int": "int",
        "xsd:float": "float",
        "xsd:double": "float",
        "xsd:boolean": "bool",
        "xsd:dateTime": "DateTime",
        "xsd:date": "Date",
        "xsd:time": "Time",
        "xsd:anyURI": "str",
        "xsd:decimal": "float",
    },
}


class TemplateEnum(Enum):
    CLASS_BASED = "panderagen_class_based"


@dataclass
class PanderaGenerator(OOCodeGenerator):
    """
    Generates Pandera python classes from a LinkML schema.

    Status: incompletely implemented

    One styles is supported:

    - panderagen_class_based
    """

    DEFAULT_TEMPLATE_PATH = "panderagen_class_based"
    DEFAULT_TEMPLATE_FILE = "pandera.jinja2"

    # ClassVars
    generatorname = os.path.basename(__file__)
    generatorstem = PurePosixPath(generatorname).stem
    generatorversion = "0.0.1"
    valid_formats = ["python"]
    file_extension = "py"
    java_style = False

    # ObjectVars
    template_file: Optional[str] = None
    template_path: str = DEFAULT_TEMPLATE_PATH

    gen_classvars: bool = True
    gen_slots: bool = True
    genmeta: bool = False
    emit_metadata: bool = True
    coerce: bool = False
    backing_form: str = "serialization"
    """specific storage format that may differ from specified inlined flags"""

    def default_value_for_type(self, typ: str) -> str:
        """Allow underlying framework to handle default if not specified."""
        return None

    @staticmethod
    def make_multivalued(range: str) -> str:
        if range == "Struct":
            return "pl.List"
        return f"List[{range}]"

    def uri_type_map(self, xsd_uri: str, template: str = None):
        if template is None:
            template = self.template_path

        return TYPEMAP[template].get(xsd_uri)

    def map_type(self, t: TypeDefinition) -> str:
        logger.info(f"type_map definition: {t}")

        typ = None

        if t.uri:
            typ = self.uri_type_map(t.uri)
            if typ is None:
                typ = self.map_type(self.schemaview.get_type(t.typeof))
        elif t.typeof:
            typ = self.map_type(self.schemaview.get_type(t.typeof))

        if typ is None:
            raise ValueError(f"{t} cannot be mapped to a type")

        return typ

    def load_template(self, template_filename):
        jinja_env = Environment(loader=PackageLoader("linkml.generators.panderagen", self.template_path))
        return jinja_env.get_template(template_filename)

    def compile_pandera(self) -> ModuleType:
        """
        Generates and compiles Pandera model
        """
        pandera_code = self.serialize()

        return compile_python(pandera_code)

    def serialize(self, rendered_module: Optional[DataframeDocument] = None) -> str:
        """
        Serialize the schema to a Pandera module as a string
        """
        if self.template_path is None:
            self.template_path = PanderaGenerator.DEFAULT_TEMPLATE_PATH

        if rendered_module is not None:
            module = rendered_module
        else:
            module = self.render()

        if self.template_file is None:
            self.template_file = PanderaGenerator.DEFAULT_TEMPLATE_FILE
        template_file = self.template_file

        template_obj = self.load_template(template_file)

        code = template_obj.render(
            doc=module,
            metamodel_version=self.schema.metamodel_version,
            model_version=self.schema.version,
            coerce=self.coerce,
            type_map=TYPEMAP,
            template_path=self.template_path,
            pandera_validator_code=None,
        )
        return code

    def render(self) -> DataframeDocument:
        """
        Create a data structure ready to pass to the serialization templates.
        """
        sv: SchemaView = self.schemaview

        module_name = camelcase(sv.schema.name)

        doc = DataframeDocument(name=module_name, package=self.package, source_schema=sv.schema)

        classes = []

        for c in self.class_handler.ordered_classes():
            cn = c.name
            safe_cn = camelcase(cn)
            annotations = {}
            identifier_or_key_slot = self.slot_handler.get_identifier_or_key_slot(cn)
            if identifier_or_key_slot:
                annotations["identifier_key_slot"] = identifier_or_key_slot.name
            ooclass = DataframeClass(
                name=safe_cn,
                description=c.description,
                package=self.package,
                fields=[],
                source_class=c,
                annotations=annotations,
            )
            classes.append(ooclass)
            if c.mixin:
                ooclass.mixin = c.mixin
            if c.mixins:
                ooclass.mixins = [(x) for x in c.mixins]
            if c.is_a:
                ooclass.is_a = self.get_class_name(c.is_a)
                parent_slots = sv.class_slots(c.is_a)
            else:
                parent_slots = []
            for sn in sv.class_slots(cn):
                oofield = self.slot_handler.handle_slot(cn, sn)
                if sn not in parent_slots:
                    ooclass.fields.append(oofield)
                ooclass.all_fields.append(oofield)

        doc.classes = classes

        return doc

    def __post_init__(self):
        super().__post_init__()
        self.slot_handler = SlotHandlerBase(self)
        self.enum_handler = EnumHandlerBase(self)
        self.class_handler = ClassHandlerBase(self)


@click.option("--package", help="Package name where relevant for generated class files")
@click.option("--template-path", help="Optional jinja2 template directory within module")
@click.option("--template-file", help="Optional jinja2 template to use for class generation")
@click.version_option(__version__, "-V", "--version")
@click.argument("yamlfile")
@click.command(name="gen-pandera")
def cli(
    yamlfile,
    package=None,
    template_path=None,
    template_file=None,
    **args,
):
    if template_path is not None and template_path not in TYPEMAP:
        raise Exception(f"Template {template_path} not supported")

    """Generate Pandera classes to represent a LinkML model"""
    gen = PanderaGenerator(
        yamlfile,
        package=package,
        template_path=template_path,
        template_file=template_file,
        **args,
    )

    print(gen.serialize())


if __name__ == "__main__":
    cli()



# Auto generated from workspaces.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-04-07T18:13:53
# Schema: workspace
#
# id: https://w3id.org/linkml/workspace
# description: A datamodel for Workspaces. A workspace is a local on-disk collection of projects managed by a
#              single user or agent
# license: https://creativecommons.org/publicdomain/zero/1.0/

from dataclasses import dataclass
from typing import Any, ClassVar, Optional, Union

from jsonasobj2 import as_dict
from linkml_runtime.linkml_model.types import String
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.utils.metamodelcore import Bool, XSDDateTime, empty_dict, empty_list
from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str
from rdflib import URIRef

metamodel_version = "1.7.0"
version = None

# Namespaces
CSVW = CurieNamespace("csvw", "http://www.w3.org/ns/csvw#")
DCAT = CurieNamespace("dcat", "http://www.w3.org/ns/dcat#")
FORMATS = CurieNamespace("formats", "http://www.w3.org/ns/formats/")
FRICTIONLESS = CurieNamespace("frictionless", "https://specs.frictionlessdata.io/")
LINKML = CurieNamespace("linkml", "https://w3id.org/linkml/")
MEDIATYPES = CurieNamespace("mediatypes", "https://www.iana.org/assignments/media-types/")
OWL = CurieNamespace("owl", "http://www.w3.org/2002/07/owl#")
PAV = CurieNamespace("pav", "http://purl.org/pav/")
RDF = CurieNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
RDFS = CurieNamespace("rdfs", "http://www.w3.org/2000/01/rdf-schema#")
SCHEMA = CurieNamespace("schema", "http://schema.org/")
SH = CurieNamespace("sh", "https://w3id.org/shacl/")
SKOS = CurieNamespace("skos", "http://www.w3.org/2004/02/skos/core#")
VOID = CurieNamespace("void", "http://rdfs.org/ns/void#")
WORKSPACE = CurieNamespace("workspace", "https://w3id.org/linkml/workspace")
XSD = CurieNamespace("xsd", "http://www.w3.org/2001/XMLSchema#")
DEFAULT_ = WORKSPACE


# Types
class FileSystemPath(String):
    """A local or absolute path on a file system"""

    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "FileSystemPath"
    type_model_uri = WORKSPACE.FileSystemPath


class ProjectName(String):
    """A project name MUST contain no whitespace and SHOULD only contains
    alphanumeric characters and hyphens (no underscores)"""

    type_class_uri = XSD.string
    type_class_curie = "xsd:string"
    type_name = "ProjectName"
    type_model_uri = WORKSPACE.ProjectName


# Class references
class ProjectName(ProjectName):
    pass


class GoogleSheetsDocId(extended_str):
    pass


Any = Any


@dataclass
class Project(YAMLRoot):
    """
    A project consists of a single root schema
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = WORKSPACE.Project
    class_class_curie: ClassVar[str] = "workspace:Project"
    class_name: ClassVar[str] = "Project"
    class_model_uri: ClassVar[URIRef] = WORKSPACE.Project

    name: Union[str, ProjectName] = None
    uuid: Optional[str] = None
    github_organization: Optional[Union[dict, "GitHubAccount"]] = None
    creation_date: Optional[Union[str, XSDDateTime]] = None
    schema: Optional[Union[dict, Any]] = None
    description: Optional[str] = None
    source_schema_path: Optional[Union[str, FileSystemPath]] = None
    data_files: Optional[Union[Union[str, FileSystemPath], list[Union[str, FileSystemPath]]]] = empty_list()
    source_google_sheet_docs: Optional[
        Union[
            dict[Union[str, GoogleSheetsDocId], Union[dict, "GoogleSheetsDoc"]],
            list[Union[dict, "GoogleSheetsDoc"]],
        ]
    ] = empty_dict()
    project_directory: Optional[Union[str, FileSystemPath]] = None
    external_project_path: Optional[Union[str, FileSystemPath]] = None
    last_saved: Optional[Union[str, XSDDateTime]] = None

    def __post_init__(self, *_: list[str], **kwargs: dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, ProjectName):
            self.name = ProjectName(self.name)

        if self.uuid is not None and not isinstance(self.uuid, str):
            self.uuid = str(self.uuid)

        if self.github_organization is not None and not isinstance(self.github_organization, GitHubAccount):
            self.github_organization = GitHubAccount(**as_dict(self.github_organization))

        if self.creation_date is not None and not isinstance(self.creation_date, XSDDateTime):
            self.creation_date = XSDDateTime(self.creation_date)

        if self.description is not None and not isinstance(self.description, str):
            self.description = str(self.description)

        if self.source_schema_path is not None and not isinstance(self.source_schema_path, FileSystemPath):
            self.source_schema_path = FileSystemPath(self.source_schema_path)

        if not isinstance(self.data_files, list):
            self.data_files = [self.data_files] if self.data_files is not None else []
        self.data_files = [v if isinstance(v, FileSystemPath) else FileSystemPath(v) for v in self.data_files]

        self._normalize_inlined_as_dict(
            slot_name="source_google_sheet_docs",
            slot_type=GoogleSheetsDoc,
            key_name="id",
            keyed=True,
        )

        if self.project_directory is not None and not isinstance(self.project_directory, FileSystemPath):
            self.project_directory = FileSystemPath(self.project_directory)

        if self.external_project_path is not None and not isinstance(self.external_project_path, FileSystemPath):
            self.external_project_path = FileSystemPath(self.external_project_path)

        if self.last_saved is not None and not isinstance(self.last_saved, XSDDateTime):
            self.last_saved = XSDDateTime(self.last_saved)

        super().__post_init__(**kwargs)


@dataclass
class GoogleSheetsDoc(YAMLRoot):
    """
    A google sheets document can contain multiple individual sheets
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = WORKSPACE.GoogleSheetsDoc
    class_class_curie: ClassVar[str] = "workspace:GoogleSheetsDoc"
    class_name: ClassVar[str] = "GoogleSheetsDoc"
    class_model_uri: ClassVar[URIRef] = WORKSPACE.GoogleSheetsDoc

    id: Union[str, GoogleSheetsDocId] = None
    sheet_ids: Optional[Union[str, list[str]]] = empty_list()

    def __post_init__(self, *_: list[str], **kwargs: dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, GoogleSheetsDocId):
            self.id = GoogleSheetsDocId(self.id)

        if not isinstance(self.sheet_ids, list):
            self.sheet_ids = [self.sheet_ids] if self.sheet_ids is not None else []
        self.sheet_ids = [v if isinstance(v, str) else str(v) for v in self.sheet_ids]

        super().__post_init__(**kwargs)


@dataclass
class GitHubAccount(YAMLRoot):
    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = WORKSPACE.GitHubAccount
    class_class_curie: ClassVar[str] = "workspace:GitHubAccount"
    class_name: ClassVar[str] = "GitHubAccount"
    class_model_uri: ClassVar[URIRef] = WORKSPACE.GitHubAccount

    username: Optional[str] = None
    password: Optional[str] = None

    def __post_init__(self, *_: list[str], **kwargs: dict[str, Any]):
        if self.username is not None and not isinstance(self.username, str):
            self.username = str(self.username)

        if self.password is not None and not isinstance(self.password, str):
            self.password = str(self.password)

        super().__post_init__(**kwargs)


@dataclass
class Workspace(YAMLRoot):
    """
    A workspace is a collection of projects managed locally on a file system
    """

    _inherited_slots: ClassVar[list[str]] = []

    class_class_uri: ClassVar[URIRef] = WORKSPACE.Workspace
    class_class_curie: ClassVar[str] = "workspace:Workspace"
    class_name: ClassVar[str] = "Workspace"
    class_model_uri: ClassVar[URIRef] = WORKSPACE.Workspace

    projects: Optional[
        Union[
            dict[Union[str, ProjectName], Union[dict, Project]],
            list[Union[dict, Project]],
        ]
    ] = empty_dict()
    github_account: Optional[Union[dict, GitHubAccount]] = None
    projects_directory: Optional[Union[str, FileSystemPath]] = None
    autosync: Optional[Union[bool, Bool]] = None

    def __post_init__(self, *_: list[str], **kwargs: dict[str, Any]):
        self._normalize_inlined_as_list(slot_name="projects", slot_type=Project, key_name="name", keyed=True)

        if self.github_account is not None and not isinstance(self.github_account, GitHubAccount):
            self.github_account = GitHubAccount(**as_dict(self.github_account))

        if self.projects_directory is not None and not isinstance(self.projects_directory, FileSystemPath):
            self.projects_directory = FileSystemPath(self.projects_directory)

        if self.autosync is not None and not isinstance(self.autosync, Bool):
            self.autosync = Bool(self.autosync)

        super().__post_init__(**kwargs)


# Enumerations


# Slots
class slots:
    pass


slots.project__name = Slot(
    uri=WORKSPACE.name,
    name="project__name",
    curie=WORKSPACE.curie("name"),
    model_uri=WORKSPACE.project__name,
    domain=None,
    range=URIRef,
)

slots.project__uuid = Slot(
    uri=WORKSPACE.uuid,
    name="project__uuid",
    curie=WORKSPACE.curie("uuid"),
    model_uri=WORKSPACE.project__uuid,
    domain=None,
    range=Optional[str],
)

slots.project__github_organization = Slot(
    uri=WORKSPACE.github_organization,
    name="project__github_organization",
    curie=WORKSPACE.curie("github_organization"),
    model_uri=WORKSPACE.project__github_organization,
    domain=None,
    range=Optional[Union[dict, GitHubAccount]],
)

slots.project__creation_date = Slot(
    uri=WORKSPACE.creation_date,
    name="project__creation_date",
    curie=WORKSPACE.curie("creation_date"),
    model_uri=WORKSPACE.project__creation_date,
    domain=None,
    range=Optional[Union[str, XSDDateTime]],
)

slots.project__schema = Slot(
    uri=WORKSPACE.schema,
    name="project__schema",
    curie=WORKSPACE.curie("schema"),
    model_uri=WORKSPACE.project__schema,
    domain=None,
    range=Optional[Union[dict, Any]],
)

slots.project__description = Slot(
    uri=WORKSPACE.description,
    name="project__description",
    curie=WORKSPACE.curie("description"),
    model_uri=WORKSPACE.project__description,
    domain=None,
    range=Optional[str],
)

slots.project__source_schema_path = Slot(
    uri=WORKSPACE.source_schema_path,
    name="project__source_schema_path",
    curie=WORKSPACE.curie("source_schema_path"),
    model_uri=WORKSPACE.project__source_schema_path,
    domain=None,
    range=Optional[Union[str, FileSystemPath]],
)

slots.project__data_files = Slot(
    uri=WORKSPACE.data_files,
    name="project__data_files",
    curie=WORKSPACE.curie("data_files"),
    model_uri=WORKSPACE.project__data_files,
    domain=None,
    range=Optional[Union[Union[str, FileSystemPath], list[Union[str, FileSystemPath]]]],
)

slots.project__source_google_sheet_docs = Slot(
    uri=WORKSPACE.source_google_sheet_docs,
    name="project__source_google_sheet_docs",
    curie=WORKSPACE.curie("source_google_sheet_docs"),
    model_uri=WORKSPACE.project__source_google_sheet_docs,
    domain=None,
    range=Optional[
        Union[
            dict[Union[str, GoogleSheetsDocId], Union[dict, GoogleSheetsDoc]],
            list[Union[dict, GoogleSheetsDoc]],
        ]
    ],
)

slots.project__project_directory = Slot(
    uri=WORKSPACE.project_directory,
    name="project__project_directory",
    curie=WORKSPACE.curie("project_directory"),
    model_uri=WORKSPACE.project__project_directory,
    domain=None,
    range=Optional[Union[str, FileSystemPath]],
)

slots.project__external_project_path = Slot(
    uri=WORKSPACE.external_project_path,
    name="project__external_project_path",
    curie=WORKSPACE.curie("external_project_path"),
    model_uri=WORKSPACE.project__external_project_path,
    domain=None,
    range=Optional[Union[str, FileSystemPath]],
)

slots.project__last_saved = Slot(
    uri=WORKSPACE.last_saved,
    name="project__last_saved",
    curie=WORKSPACE.curie("last_saved"),
    model_uri=WORKSPACE.project__last_saved,
    domain=None,
    range=Optional[Union[str, XSDDateTime]],
)

slots.googleSheetsDoc__id = Slot(
    uri=WORKSPACE.id,
    name="googleSheetsDoc__id",
    curie=WORKSPACE.curie("id"),
    model_uri=WORKSPACE.googleSheetsDoc__id,
    domain=None,
    range=URIRef,
)

slots.googleSheetsDoc__sheet_ids = Slot(
    uri=WORKSPACE.sheet_ids,
    name="googleSheetsDoc__sheet_ids",
    curie=WORKSPACE.curie("sheet_ids"),
    model_uri=WORKSPACE.googleSheetsDoc__sheet_ids,
    domain=None,
    range=Optional[Union[str, list[str]]],
)

slots.gitHubAccount__username = Slot(
    uri=WORKSPACE.username,
    name="gitHubAccount__username",
    curie=WORKSPACE.curie("username"),
    model_uri=WORKSPACE.gitHubAccount__username,
    domain=None,
    range=Optional[str],
)

slots.gitHubAccount__password = Slot(
    uri=WORKSPACE.password,
    name="gitHubAccount__password",
    curie=WORKSPACE.curie("password"),
    model_uri=WORKSPACE.gitHubAccount__password,
    domain=None,
    range=Optional[str],
)

slots.workspace__projects = Slot(
    uri=WORKSPACE.projects,
    name="workspace__projects",
    curie=WORKSPACE.curie("projects"),
    model_uri=WORKSPACE.workspace__projects,
    domain=None,
    range=Optional[
        Union[
            dict[Union[str, ProjectName], Union[dict, Project]],
            list[Union[dict, Project]],
        ]
    ],
)

slots.workspace__github_account = Slot(
    uri=WORKSPACE.github_account,
    name="workspace__github_account",
    curie=WORKSPACE.curie("github_account"),
    model_uri=WORKSPACE.workspace__github_account,
    domain=None,
    range=Optional[Union[dict, GitHubAccount]],
)

slots.workspace__projects_directory = Slot(
    uri=WORKSPACE.projects_directory,
    name="workspace__projects_directory",
    curie=WORKSPACE.curie("projects_directory"),
    model_uri=WORKSPACE.workspace__projects_directory,
    domain=None,
    range=Optional[Union[str, FileSystemPath]],
)

slots.workspace__autosync = Slot(
    uri=WORKSPACE.autosync,
    name="workspace__autosync",
    curie=WORKSPACE.curie("autosync"),
    model_uri=WORKSPACE.workspace__autosync,
    domain=None,
    range=Optional[Union[bool, Bool]],
)



import copy
from datetime import datetime
from pathlib import Path
from typing import Optional, TextIO, Union
from urllib.parse import urlparse

import yaml
from dateutil.parser import ParserError, parse
from hbreader import FileInfo, HBType, detect_type
from linkml_runtime.linkml_model.meta import SchemaDefinition, metamodel_version
from linkml_runtime.loaders import yaml_loader
from linkml_runtime.utils.yamlutils import YAMLMark, YAMLRoot

from linkml.utils.mergeutils import set_from_schema

DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S"
yaml.error.Mark = YAMLMark


# Override the default linkml missing value tests
def mrf(self, field_name: str) -> None:
    if isinstance(self, SchemaDefinition) and field_name == "name" and self.id:
        id_parts = self.id.replace("#", "/").rsplit("/")
        self.name = id_parts[-1]
    else:
        YAMLRoot.MissingRequiredField(self, f"{type(self).__name__}.{field_name}")


SchemaDefinition.MissingRequiredField = mrf


def load_raw_schema(
    data: Union[str, dict, TextIO, Path],
    source_file: Optional[str] = None,
    source_file_date: Optional[str] = None,
    source_file_size: Optional[int] = None,
    base_dir: Optional[str] = None,
    merge_modules: Optional[bool] = True,
    emit_metadata: Optional[bool] = True,
) -> SchemaDefinition:
    """Load and flatten SchemaDefinition from a file name, a URL or a block of text

    @param data: URL, file name or block of text YAML Object or open file handle
    @param source_file: Source file name for the schema if data is type TextIO
    @param source_file_date: timestamp of source file if data is type TextIO
    @param source_file_size: size of source file if data is type TextIO
    @param base_dir: Working directory or base URL of sources
    @param merge_modules: True means combine modules into one source, false means keep separate
    @param emit_metadata: True means add source file info to the output
    @return: Un-processed Schema Definition object
    """

    def _name_from_url(url) -> str:
        return urlparse(url).path.rsplit("/", 1)[-1].rsplit(".", 1)[0]

    # Passing a URL or file name
    if detect_type(data, base_dir) not in (HBType.STRING, HBType.STRINGABLE):
        assert source_file is None, "source_file parameter not allowed if data is a file or URL"
        assert source_file_date is None, "source_file_date parameter not allowed if data is a file or URL"
        assert source_file_size is None, "source_file_size parameter not allowed if data is a file or URL"

    if isinstance(data, Path):
        data = str(data)

    # Convert the input into a valid SchemaDefinition
    if isinstance(data, (str, dict, TextIO)):
        # TODO: Build a generic loader that detects type from suffix or content and invokes the appropriate loader
        schema_metadata = FileInfo()
        schema_metadata.source_file = source_file
        schema_metadata.source_file_date = source_file_date
        schema_metadata.source_file_size = source_file_size
        schema_metadata.base_path = base_dir
        schema = yaml_loader.load(
            copy.deepcopy(data) if isinstance(data, dict) else data,
            SchemaDefinition,
            base_dir=base_dir,
            metadata=schema_metadata,
        )
    elif isinstance(data, SchemaDefinition):
        schema = copy.deepcopy(data)
    else:
        raise ValueError("Unrecognized input to raw loader")

    if schema is None:
        raise ValueError("Empty schema - cannot process")

    if schema.name is None:
        if schema.id is None:
            raise ValueError("Unable to determine schema name")
        else:
            schema.name = _name_from_url(schema.id)
    elif schema.id is None:
        # TODO: figure out how to generate this from the default_prefix and namespace map
        raise ValueError("Schema identifier must be supplied")

    if emit_metadata:
        schema.source_file = schema_metadata.source_file
        src_date = schema_metadata.source_file_date
        try:
            schema.source_file_date = parse(src_date).strftime(DATETIME_FORMAT) if src_date else None
        except ParserError:
            schema.source_file_date = src_date
        schema.source_file_size = schema_metadata.source_file_size
        schema.generation_date = datetime.now().strftime(DATETIME_FORMAT)
    schema.metamodel_version = metamodel_version

    set_from_schema(schema)

    return schema
